from __future__ import annotations

from abc import ABC
from collections.abc import Iterable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, cast

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
    CogniteResponse,
    IdTransformerMixin,
    InternalIdTransformerMixin,
    NameTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.capabilities import Capability, ProjectCapabilityList
from cognite.client.utils._importing import local_import

if TYPE_CHECKING:
    import pandas as pd

    from cognite.client import AsyncCogniteClient

ALL_USER_ACCOUNTS = "allUserAccounts"


@dataclass
class GroupAttributesToken(CogniteObject):
    """List of applications (represented by their application ID) this group is valid for"""

    app_ids: list[str] = field(default_factory=list)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(app_ids=resource.get("appIds", []))


@dataclass
class GroupAttributes(CogniteObject):
    """Attributes derived from access token"""

    token: GroupAttributesToken | None = None
    _unknown_properties: dict[str, Any] = field(default_factory=dict, init=False, repr=False)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dumps the attributes to a dictionary"""
        dumped = super().dump(camel_case=camel_case)
        if self.token is not None:
            dumped["token"] = self.token.dump(camel_case=camel_case)
        if self._unknown_properties:
            dumped.update(self._unknown_properties)
        return dumped

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        token: GroupAttributesToken | None = None
        if "token" in resource:
            token = GroupAttributesToken._load(resource["token"], cognite_client=cognite_client)
        instance = cls(token=token)
        existing = {"token"}
        instance._unknown_properties = {key: value for key, value in resource.items() if key not in existing}
        return instance


class GroupCore(WriteableCogniteResource["GroupWrite"], ABC):
    """No description.

    Args:
        name (str): Name of the group.
        source_id (str | None): ID of the group in the source. If this is the same ID as a group in the IdP, a service account in that group will implicitly be a part of this group as well. Can not be used together with 'members'.
        capabilities (list[Capability] | Capability | None): List of capabilities (acls) this group should grant its users.
        attributes (GroupAttributes | None): Attributes of the group, this scopes down access based on the attributes specified.
        metadata (dict[str, str] | None): Custom, immutable application specific metadata. String key -> String value. Limits: Key are at most 32 bytes. Values are at most 512 bytes. Up to 16 key-value pairs. Total size is at most 4096.
        members (Literal['allUserAccounts'] | list[str] | None): Specifies which users are members of the group. Can not be used together with 'source_id'.
    """

    def __init__(
        self,
        name: str,
        source_id: str | None,
        capabilities: list[Capability] | Capability | None,
        attributes: GroupAttributes | None,
        metadata: dict[str, str] | None,
        members: Literal["allUserAccounts"] | list[str] | None,
    ) -> None:
        self.name = name
        self.source_id = source_id
        self.capabilities = [capabilities] if isinstance(capabilities, Capability) else capabilities
        self.attributes = attributes
        self.metadata = metadata
        self.members = members

    @classmethod
    def _load(
        cls, resource: dict, cognite_client: AsyncCogniteClient | None = None, allow_unknown: bool = False
    ) -> Self:
        return cls(
            name=resource["name"],
            source_id=resource.get("sourceId"),
            capabilities=[Capability.load(c, allow_unknown=allow_unknown) for c in resource.get("capabilities", [])]
            or None,
            attributes=(
                GroupAttributes._load(resource["attributes"], cognite_client=cognite_client)
                if isinstance(resource.get("attributes"), dict)
                else None
            ),
            metadata=resource.get("metadata"),
            members=resource.get("members"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        dumped = super().dump(camel_case=camel_case)
        if self.capabilities is not None:
            dumped["capabilities"] = [c.dump(camel_case=camel_case) for c in self.capabilities]
        if self.attributes is not None:
            dumped["attributes"] = self.attributes.dump(camel_case=camel_case)
        return dumped


class Group(GroupCore):
    """Groups are used to give principals the capabilities to access CDF resources. One principal
    can be a member in multiple groups and one group can have multiple members.

    Groups can either be managed through the external identity provider for the project or managed by CDF.

    Args:
        id (int): No description.
        name (str): Name of the group.
        source_id (str | None): ID of the group in the source. If this is the same ID as a group in the IdP, a service account in that group will implicitly be a part of this group as well. Can not be used together with 'members'.
        capabilities (list[Capability] | Capability | None): List of capabilities (acls) this group should grant its users.
        attributes (GroupAttributes | None): Attributes of the group, this scopes down access based on the attributes specified.
        is_deleted (bool | None): No description.
        deleted_time (int | None): No description.
        metadata (dict[str, str] | None): Custom, immutable application specific metadata. String key -> String value. Limits: Key are at most 32 bytes. Values are at most 512 bytes. Up to 16 key-value pairs. Total size is at most 4096.
        members (Literal['allUserAccounts'] | list[str] | None): Specifies which users are members of the group. Can not be used together with 'source_id'.
        cognite_client (AsyncCogniteClient | None): No description.
    """

    def __init__(
        self,
        id: int,
        name: str,
        source_id: str | None,
        capabilities: list[Capability] | Capability | None,
        attributes: GroupAttributes | None,
        is_deleted: bool | None,
        deleted_time: int | None,
        metadata: dict[str, str] | None = None,
        members: Literal["allUserAccounts"] | list[str] | None = None,
        cognite_client: AsyncCogniteClient | None = None,
    ) -> None:
        super().__init__(
            name=name,
            source_id=source_id,
            capabilities=capabilities,
            attributes=attributes,
            metadata=metadata,
            members=members,
        )
        self.id: int = id
        self.is_deleted = is_deleted
        self.deleted_time = deleted_time
        self._cognite_client = cast("AsyncCogniteClient", cognite_client)

    def as_write(self) -> GroupWrite:
        """Returns a writing version of this group."""
        return GroupWrite(
            name=self.name,
            source_id=self.source_id,
            capabilities=self.capabilities,
            attributes=self.attributes,
            metadata=self.metadata,
            members=self.members,
        )

    @property
    def is_externally_managed(self) -> bool:
        # Note: `source_id` may be returned as the empty string, hence the 'truthy' test
        if self.source_id and self.members is not None:
            raise ValueError("Managed status is unknown")
        return self.source_id is not None

    @property
    def is_managed_in_cdf(self) -> bool:
        if self.source_id and self.members is not None:
            raise ValueError("Managed status is unknown")
        return self.members is not None

    @classmethod
    def _load(
        cls, resource: dict, cognite_client: AsyncCogniteClient | None = None, allow_unknown: bool = False
    ) -> Group:
        return cls(
            id=resource["id"],
            name=resource["name"],
            source_id=resource.get("sourceId"),
            attributes=(attrs := resource.get("attributes")) and GroupAttributes._load(attrs, cognite_client),
            capabilities=[Capability.load(c, allow_unknown) for c in resource.get("capabilities", [])] or None,
            is_deleted=resource.get("isDeleted"),
            deleted_time=resource.get("deletedTime"),
            metadata=resource.get("metadata"),
            members=resource.get("members"),
            cognite_client=cognite_client,
        )

    def to_pandas(
        self,
        expand_metadata: bool = False,
        metadata_prefix: str = "metadata.",
        ignore: list[str] | None = None,
        camel_case: bool = False,
        convert_timestamps: bool = True,
    ) -> pd.DataFrame:
        df = super().to_pandas(expand_metadata, metadata_prefix, ignore, camel_case, convert_timestamps)

        # The API uses -1 to represent "no deleted time". It looks weird if deleted = False,
        # but deleted_time = 1969-12-31 23:59:59.999:
        key = "deletedTime" if camel_case else "deleted_time"
        if self.deleted_time == -1 and convert_timestamps and key in df.index:
            df.at[key, "value"] = local_import("pandas").NaT
        return df


class GroupWrite(GroupCore):
    """Groups are used to give principals the capabilities to access CDF resources. One principal
    can be a member in multiple groups and one group can have multiple members.

    Groups can either be managed through the external identity provider for the project or managed by CDF.

    Args:
        name (str): Name of the group.
        source_id (str | None): ID of the group in the source. If this is the same ID as a group in the IdP, a service account in that group will implicitly be a part of this group as well. Can not be used together with 'members'.
        capabilities (list[Capability] | None): List of capabilities (acls) this group should grant its users.
        attributes (GroupAttributes | None): Attributes of the group, this scopes down access based on the attributes specified.
        metadata (dict[str, str] | None): Custom, immutable application specific metadata. String key -> String value. Limits: Key are at most 32 bytes. Values are at most 512 bytes. Up to 16 key-value pairs. Total size is at most 4096.
        members (Literal['allUserAccounts'] | list[str] | None): Specifies which users are members of the group. Can not be used together with 'source_id'.
    """

    def __init__(
        self,
        name: str,
        source_id: str | None = None,
        capabilities: list[Capability] | None = None,
        attributes: GroupAttributes | None = None,
        metadata: dict[str, str] | None = None,
        members: Literal["allUserAccounts"] | list[str] | None = None,
    ) -> None:
        super().__init__(
            name=name,
            source_id=source_id,
            attributes=attributes,
            capabilities=capabilities,
            metadata=metadata,
            members=members,
        )

    def as_write(self) -> GroupWrite:
        """Returns this GroupWrite instance."""
        return self


class GroupWriteList(CogniteResourceList[GroupWrite], NameTransformerMixin):
    _RESOURCE = GroupWrite

    @classmethod
    def _load(
        cls,
        resource_list: Iterable[dict[str, Any]],
        cognite_client: AsyncCogniteClient | None = None,
        allow_unknown: bool = False,
    ) -> Self:
        return cls(
            [cls._RESOURCE._load(res, cognite_client, allow_unknown) for res in resource_list],
            cognite_client=cognite_client,
        )


class GroupList(WriteableCogniteResourceList[GroupWrite, Group], NameTransformerMixin, InternalIdTransformerMixin):
    _RESOURCE = Group

    @classmethod
    def _load(
        cls,
        resource_list: Iterable[dict[str, Any]],
        cognite_client: AsyncCogniteClient | None = None,
        allow_unknown: bool = False,
    ) -> Self:
        return cls(
            [cls._RESOURCE._load(res, cognite_client, allow_unknown) for res in resource_list],
            cognite_client=cognite_client,
        )

    def as_write(self) -> GroupWriteList:
        """Returns a writing version of this group list."""
        return GroupWriteList([s.as_write() for s in self], cognite_client=self._get_cognite_client())

    def to_pandas(
        self,
        camel_case: bool = False,
        expand_metadata: bool = False,
        metadata_prefix: str = "metadata.",
        convert_timestamps: bool = True,
    ) -> pd.DataFrame:
        df = super().to_pandas(camel_case, expand_metadata, metadata_prefix, convert_timestamps)

        # The API uses -1 to represent "no deleted time". It looks weird if deleted = False,
        # but deleted_time = 1969-12-31 23:59:59.999:
        key = "deletedTime" if camel_case else "deleted_time"
        if convert_timestamps and key in df:
            pd = local_import("pandas")
            df.loc[df[key] == pd.Timestamp(-1, unit="ms"), key] = pd.NaT
        return df


class SecurityCategoryCore(WriteableCogniteResource["SecurityCategoryWrite"], ABC):
    """No description.

    Args:
        name (str | None): Name of the security category
    """

    def __init__(self, name: str | None = None) -> None:
        self.name = name


class SecurityCategory(SecurityCategoryCore):
    """Security categories can be used to restrict access to a resource.
    This is the reading version of a security category, which is used when retrieving security categories.

    Args:
        id (int): Id of the security category
        name (str | None): Name of the security category
        cognite_client (AsyncCogniteClient | None): The client to associate with this object.
    """

    def __init__(self, id: int, name: str | None, cognite_client: AsyncCogniteClient | None = None) -> None:
        super().__init__(name=name)
        self.id = id
        self._cognite_client = cast("AsyncCogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(id=resource["id"], name=resource.get("name"), cognite_client=cognite_client)

    def as_write(self) -> SecurityCategoryWrite:
        """Returns a writing version of this security category."""
        if self.name is None:
            raise ValueError("SecurityCategory must have an id to be used as write")
        return SecurityCategoryWrite(name=self.name)


class SecurityCategoryWrite(SecurityCategoryCore):
    """Security categories can be used to restrict access to a resource.
    This is the writing version of a security category, which is used when creating security categories.


    Args:
        name (str): Name of the security category
    """

    def __init__(self, name: str) -> None:
        super().__init__(name=name)

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(name=resource["name"])

    def as_write(self) -> SecurityCategoryWrite:
        """Returns this SecurityCategoryWrite instance."""
        return self


class SecurityCategoryWriteList(CogniteResourceList[SecurityCategoryWrite], NameTransformerMixin):
    _RESOURCE = SecurityCategoryWrite


class SecurityCategoryList(
    WriteableCogniteResourceList[SecurityCategoryWrite, SecurityCategory],
    InternalIdTransformerMixin,
    NameTransformerMixin,
):
    _RESOURCE = SecurityCategory

    def as_write(self) -> SecurityCategoryWriteList:
        """Returns a writing version of this security category list."""
        return SecurityCategoryWriteList([s.as_write() for s in self], cognite_client=self._get_cognite_client())


class ProjectSpec(CogniteResponse):
    """A CDF project spec

    Args:
        url_name (str): The url name for the project
        groups (list[int]): Group ids in the project
    """

    def __init__(self, url_name: str, groups: list[int]) -> None:
        self.url_name = url_name
        self.groups = groups

    @property
    def project_url_name(self) -> str:
        return self.url_name

    @classmethod
    def load(cls, api_response: dict[str, Any]) -> ProjectSpec:
        return cls(url_name=api_response["projectUrlName"], groups=api_response["groups"])

    def dump(self, camel_case: bool = True) -> dict[str, str | list[int]]:
        return {
            "projectUrlName" if camel_case else "project_url_name": self.url_name,
            "groups": self.groups,
        }


class TokenInspection(CogniteResponse):
    """Current login status

    Args:
        subject (str): Subject (sub claim) of JWT.
        projects (list[ProjectSpec]): Projects this token is valid for.
        capabilities (ProjectCapabilityList): Capabilities associated with this token.
    """

    def __init__(self, subject: str, projects: list[ProjectSpec], capabilities: ProjectCapabilityList) -> None:
        self.subject = subject
        self.projects = projects
        self.capabilities = capabilities

    @classmethod
    def load(
        cls,
        api_response: dict[str, Any],
        cognite_client: AsyncCogniteClient | None = None,
        allow_unknown: bool = False,
    ) -> TokenInspection:
        return cls(
            subject=api_response["subject"],
            projects=[ProjectSpec.load(p) for p in api_response["projects"]],
            capabilities=ProjectCapabilityList._load(api_response["capabilities"], cognite_client, allow_unknown),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "subject": self.subject,
            "projects": [p.dump(camel_case=camel_case) for p in self.projects],
            "capabilities": self.capabilities.dump(camel_case=camel_case),
        }


SessionStatus: TypeAlias = Literal["READY", "ACTIVE", "CANCELLED", "EXPIRED", "REVOKED", "ACCESS_LOST"]
SessionType: TypeAlias = Literal["CLIENT_CREDENTIALS", "TOKEN_EXCHANGE", "ONESHOT_TOKEN_EXCHANGE"]


class CreatedSession(CogniteResponse):
    """Session creation related information

    Args:
        id (int): ID of the created session.
        status (SessionStatus): Current status of the session.
        nonce (str): Nonce to be passed to the internal service that will bind the session
        type (SessionType | None): Credentials kind used to create the session.
        client_id (str | None): Client ID in identity provider. Returned only if the session was created using client credentials
    """

    def __init__(
        self,
        id: int,
        status: SessionStatus,
        nonce: str,
        type: SessionType | None = None,
        client_id: str | None = None,
    ) -> None:
        self.id = id
        self.status = status
        self.nonce = nonce
        self.type = type
        self.client_id = client_id

    @classmethod
    def load(cls, response: dict[str, Any]) -> CreatedSession:
        return cls(
            id=response["id"],
            status=response["status"],
            nonce=response["nonce"],
            type=response.get("type"),
            client_id=response.get("clientId"),
        )


class Session(CogniteResource):
    """Session status

    Args:
        id (int): ID of the session.
        type (SessionType): Credentials kind used to create the session.
        status (SessionStatus): Current status of the session.
        creation_time (int): Session creation time, in milliseconds since 1970
        expiration_time (int): Session expiry time, in milliseconds since 1970. This value is updated on refreshing a token
        client_id (str | None): Client ID in identity provider. Returned only if the session was created using client credentials
        cognite_client (AsyncCogniteClient | None): No description.
    """

    def __init__(
        self,
        id: int,
        type: SessionType,
        status: SessionStatus,
        creation_time: int,
        expiration_time: int,
        client_id: str | None,
        cognite_client: AsyncCogniteClient | None = None,
    ) -> None:
        self.id = id
        self.type = type
        self.status = status
        self.creation_time = creation_time
        self.expiration_time = expiration_time
        self.client_id = client_id

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            id=resource["id"],
            type=resource["type"],
            status=resource["status"],
            creation_time=resource["creationTime"],
            expiration_time=resource["expirationTime"],
            client_id=resource.get("clientId"),
            cognite_client=cognite_client,
        )


class SessionList(CogniteResourceList[Session], IdTransformerMixin):
    _RESOURCE = Session


class ClientCredentials(CogniteResource):
    """Client credentials for session creation

    Args:
        client_id (str): Client ID from identity provider.
        client_secret (str): Client secret from identity provider.
    """

    def __init__(self, client_id: str, client_secret: str) -> None:
        self.client_id = client_id
        self.client_secret = client_secret

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "clientId" if camel_case else "client_id": self.client_id,
            "clientSecret" if camel_case else "client_secret": self.client_secret,
        }

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> ClientCredentials:
        return cls(client_id=resource["clientId"], client_secret=resource["clientSecret"])



"""
===============================================================================
506bda1e5a8fa5d128a4da3ae05bb18b
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from cognite.client import AsyncCogniteClient
from cognite.client._sync_api.postgres_gateway.tables import SyncTablesAPI
from cognite.client._sync_api.postgres_gateway.users import SyncUsersAPI
from cognite.client._sync_api_client import SyncAPIClient

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class SyncPostgresGatewaysAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client
        self.users = SyncUsersAPI(async_client)
        self.tables = SyncTablesAPI(async_client)



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, ClassVar, Literal

from cognite.client.data_classes._base import CogniteObject, CogniteResource, CogniteResourceList
from cognite.client.utils._text import convert_all_keys_to_camel_case

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


@dataclass
class MessageContent(CogniteObject, ABC):
    """Base class for message content types."""

    _type: ClassVar[str]  # To be set by concrete classes

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> MessageContent:
        """Dispatch to the correct concrete content class based on `type`."""
        content_type = data.get("type", "")
        content_class = _MESSAGE_CONTENT_CLS_BY_TYPE.get(content_type, UnknownContent)
        # Delegate to the concrete class' loader
        return content_class._load_content(data)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the content to a dictionary."""
        output = super().dump(camel_case=camel_case)
        output["type"] = self._type
        return output

    @classmethod
    @abstractmethod
    def _load_content(cls, data: dict[str, Any]) -> MessageContent:
        """Create a concrete content instance from raw data."""
        ...


@dataclass
class TextContent(MessageContent):
    """Text content for messages.

    Args:
        text (str): The text content.
    """

    _type: ClassVar[str] = "text"
    text: str = ""

    @classmethod
    def _load_content(cls, data: dict[str, Any]) -> TextContent:
        return cls(text=data["text"])


@dataclass
class UnknownContent(MessageContent):
    """Unknown content type for forward compatibility.

    Args:
        data (dict[str, Any]): The raw content data.
        type (str): The content type.
    """

    type: str
    data: dict[str, Any] = field(default_factory=dict)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result = self.data.copy()
        result["type"] = self.type
        return result

    @classmethod
    def _load_content(cls, data: dict[str, Any]) -> UnknownContent:
        content_type = data.get("type", "")
        return cls(data=data, type=content_type)


# Build the mapping AFTER concrete classes are defined
_MESSAGE_CONTENT_CLS_BY_TYPE: dict[str, type[MessageContent]] = {
    subclass._type: subclass  # type: ignore[type-abstract]
    for subclass in MessageContent.__subclasses__()
    if hasattr(subclass, "_type") and not getattr(subclass, "__abstractmethods__", None)
}


@dataclass(frozen=True, slots=True)
class Action(CogniteObject, ABC):
    """Base class for all action types that can be provided to an agent."""

    _type: ClassVar[str]

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Action:
        """Dispatch to the correct concrete action class based on `type`."""
        action_type = data.get("type", "")
        action_class = _ACTION_CLS_BY_TYPE.get(action_type, UnknownAction)
        return action_class._load_action(data, cognite_client)

    @classmethod
    @abstractmethod
    def _load_action(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Action:
        """Create a concrete action instance from raw data."""
        ...


@dataclass(frozen=True, slots=True)
class ClientToolAction(Action):
    """A client-side tool definition that can be called by the agent.

    Args:
        name (str): The name of the client tool to call.
        description (str): A description of what the function does. The language model will use this description when selecting the function and interpreting its parameters.
        parameters (dict[str, object]): The parameters the function accepts, described as a JSON Schema object.
    """

    _type: ClassVar[str] = "clientTool"
    name: str
    description: str
    parameters: dict[
        str, object
    ]  # TODO: We do not want the user to have to handle this, instead e.g. set json schema from Pydantic classes, or function signatures, like ClientToolAction.from_function(func: Callable)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "type": self._type,
            self._type: {
                "name": self.name,
                "description": self.description,
                "parameters": self.parameters,
            },
        }

    @classmethod
    def _load_action(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ClientToolAction:
        client_tool = data[cls._type]
        return cls(
            name=client_tool["name"],
            description=client_tool["description"],
            parameters=client_tool["parameters"],
        )


@dataclass(frozen=True, slots=True)
class UnknownAction(Action):
    """Unknown action type for forward compatibility.

    Args:
        type (str): The action type.
        data (dict[str, object]): The raw action data.
    """

    type: str
    data: dict[str, object] = field(default_factory=dict)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result = self.data.copy()
        result["type"] = self.type
        return result

    @classmethod
    def _load_action(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> UnknownAction:
        action_type = data.get("type", "")
        return cls(data=data, type=action_type)


# Build the mapping AFTER concrete classes are defined
_ACTION_CLS_BY_TYPE: dict[str, type[Action]] = {
    subclass._type: subclass  # type: ignore[type-abstract]
    for subclass in Action.__subclasses__()
    if subclass is not UnknownAction and hasattr(subclass, "_type")
}


@dataclass(frozen=True, slots=True)
class ActionCall(CogniteObject, ABC):
    """Base class for action calls requested by the agent."""

    _type: ClassVar[str]
    action_id: str

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ActionCall:
        """Dispatch to the correct concrete action call class based on `type`."""
        action_type = data.get("type", "")
        action_class = _ACTION_CALL_CLS_BY_TYPE.get(action_type, UnknownActionCall)
        return action_class._load_call(data, cognite_client)

    @classmethod
    @abstractmethod
    def _load_call(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ActionCall:
        """Create a concrete action call instance from raw data."""
        ...


@dataclass(frozen=True, slots=True)
class ClientToolCall(ActionCall):
    """A client tool call requested by the agent.

    Args:
        action_id (str): The unique identifier for this action call.
        name (str): The name of the client tool being called.
        arguments (dict[str, object]): The parsed arguments for the tool call.
    """

    _type: ClassVar[str] = "clientTool"
    action_id: str
    name: str
    arguments: dict[str, object]  # Always a json string, in line with OpenAI's API pattern

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "type": self._type,
            "actionId" if camel_case else "action_id": self.action_id,
            self._type: {
                "name": self.name,
                "arguments": json.dumps(self.arguments),
            },
        }

    @classmethod
    def _load_call(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ClientToolCall:
        client_tool = data[cls._type]
        arguments_str = client_tool["arguments"]
        return cls(
            action_id=data["actionId"],
            name=client_tool["name"],
            arguments=json.loads(arguments_str),
        )


@dataclass(frozen=True, slots=True)
class ToolConfirmationCall(ActionCall):
    """A tool confirmation request from the agent.

    Args:
        action_id (str): The unique identifier for this action call.
        content (MessageContent): The confirmation message content.
        tool_name (str): The name of the tool requiring confirmation.
        tool_arguments (dict[str, object]): The arguments for the tool call.
        tool_description (str): Description of what the tool does.
        tool_type (str): The type of tool (e.g., "runPythonCode", "callRestApi").
        details (dict[str, object] | None): Optional additional details about the tool call.
    """

    _type: ClassVar[str] = "toolConfirmation"
    action_id: str
    content: MessageContent
    tool_name: str
    tool_arguments: dict[str, object]
    tool_description: str
    tool_type: str
    details: dict[str, object] | None = None

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result: dict[str, Any] = {
            "type": self._type,
            "actionId" if camel_case else "action_id": self.action_id,
            self._type: {
                "content": self.content.dump(camel_case=camel_case),
                "toolName" if camel_case else "tool_name": self.tool_name,
                "toolArguments" if camel_case else "tool_arguments": self.tool_arguments,
                "toolDescription" if camel_case else "tool_description": self.tool_description,
                "toolType" if camel_case else "tool_type": self.tool_type,
            },
        }
        if self.details is not None:
            result[self._type]["details"] = self.details
        return result

    @classmethod
    def _load_call(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ToolConfirmationCall:
        tool_confirmation = data[cls._type]
        content = MessageContent._load(tool_confirmation["content"])
        return cls(
            action_id=data["actionId"],
            content=content,
            tool_name=tool_confirmation["toolName"],
            tool_arguments=tool_confirmation["toolArguments"],
            tool_description=tool_confirmation["toolDescription"],
            tool_type=tool_confirmation["toolType"],
            details=tool_confirmation.get("details"),
        )


@dataclass(frozen=True, slots=True)
class UnknownActionCall(ActionCall):
    """Unknown action call type for forward compatibility.

    Args:
        action_id (str): The unique identifier for this action call.
        type (str): The action call type.
        data (dict[str, object]): The raw action call data.
    """

    action_id: str
    type: str
    data: dict[str, object] = field(default_factory=dict)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result = self.data.copy()
        result["type"] = self.type
        result["actionId" if camel_case else "action_id"] = self.action_id
        return result

    @classmethod
    def _load_call(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> UnknownActionCall:
        action_type = data["type"]
        action_id = data["actionId"]
        return cls(action_id=action_id, data=data, type=action_type)


# Build the mapping AFTER concrete classes are defined
_ACTION_CALL_CLS_BY_TYPE: dict[str, type[ActionCall]] = {
    subclass._type: subclass  # type: ignore[type-abstract]
    for subclass in ActionCall.__subclasses__()
    if subclass is not UnknownActionCall and hasattr(subclass, "_type")
}


@dataclass
class Message(CogniteResource):
    """A message to send to an agent.

    Args:
        content (str | MessageContent): The message content. If a string is provided,
            it will be converted to TextContent.
        role (Literal["user"]): The role of the message sender. Defaults to "user".
    """

    content: MessageContent
    role: Literal["user"] = "user"

    def __init__(self, content: str | MessageContent, role: Literal["user"] = "user") -> None:
        if isinstance(content, str):
            self.content = TextContent(text=content)
        else:
            self.content = content
        self.role = role

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "content": self.content.dump(camel_case=camel_case),
            "role": self.role,
        }

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Message:
        content = MessageContent._load(data["content"])
        return cls(content=content, role=data["role"])


class MessageList(CogniteResourceList[Message]):
    """List of messages."""

    _RESOURCE = Message


@dataclass(frozen=True, slots=True)
class ActionResult(CogniteObject, ABC):
    """Base class for action execution results."""

    _type: ClassVar[str]
    _role: ClassVar[Literal["action"]] = "action"
    action_id: str


@dataclass(frozen=True, slots=True)
class ClientToolResult(ActionResult):
    """Result of executing a client tool, for sending back to the agent.

    Args:
        action_id (str): The ID of the action being responded to.
        content (str | MessageContent): The result of executing the action.
        data (list[Any] | None): Optional structured data.
    """

    _type: ClassVar[str] = "clientTool"
    content: MessageContent
    data: list[Any] | None = None

    def __init__(self, action_id: str, content: str | MessageContent, data: list[Any] | None = None) -> None:
        object.__setattr__(self, "action_id", action_id)
        object.__setattr__(self, "content", TextContent(text=content) if isinstance(content, str) else content)
        object.__setattr__(self, "data", data)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "role": self._role,
            "type": self._type,
            "actionId" if camel_case else "action_id": self.action_id,
            "content": self.content.dump(camel_case=camel_case),
            "data": self.data or [],
        }

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ClientToolResult:
        """Load from dumped data. Only used for testing."""
        content = MessageContent._load(data["content"], cognite_client)
        return cls(
            action_id=data["actionId"],
            content=content,
            data=data.get("data"),
        )


@dataclass(frozen=True, slots=True)
class ToolConfirmationResult(ActionResult):
    """Result of a tool confirmation request.

    Args:
        action_id (str): The ID of the action being responded to.
        status (Literal["ALLOW", "DENY"]): Whether to allow or deny the tool execution.
    """

    _type: ClassVar[str] = "toolConfirmation"
    status: Literal["ALLOW", "DENY"]

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "role": self._role,
            "type": self._type,
            "actionId" if camel_case else "action_id": self.action_id,
            "status": self.status,
        }

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ToolConfirmationResult:
        """Load from dumped data. Not used to load from API response."""
        return cls(
            action_id=data.get("actionId", data.get("action_id", "")),
            status=data["status"],
        )


@dataclass
class AgentDataItem(CogniteObject):
    """Data item in agent response.

    Args:
        type (str): The type of data item.
        data (dict[str, Any]): The data payload.
    """

    type: str
    data: dict[str, Any]

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result = {"type": self.type}
        if self.data:
            result.update(self.data if not camel_case else convert_all_keys_to_camel_case(self.data))
        return result

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> AgentDataItem:
        item_type = data["type"]
        item_data = {k: v for k, v in data.items() if k != "type"}
        return cls(type=item_type, data=item_data)


@dataclass
class AgentReasoningItem(CogniteObject):
    """Reasoning item in agent response.

    Args:
        content (list[MessageContent]): The reasoning content.
    """

    content: list[MessageContent]

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "content": [item.dump(camel_case=camel_case) for item in self.content],
        }

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> AgentReasoningItem:
        content = [MessageContent._load(item) for item in data.get("content", [])]
        return cls(content=content)


@dataclass
class AgentMessage(CogniteResource):
    """A message from an agent.

    Args:
        content (MessageContent | None): The message content.
        data (list[AgentDataItem] | None): Data items in the response.
        reasoning (list[AgentReasoningItem] | None): Reasoning items in the response.
        actions (list[ActionCall] | None): Action calls requested by the agent.
        role (Literal["agent"]): The role of the message sender.
    """

    content: MessageContent | None = None
    data: list[AgentDataItem] | None = None
    reasoning: list[AgentReasoningItem] | None = None
    actions: list[ActionCall] | None = None
    role: Literal["agent"] = "agent"

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result: dict[str, Any] = {"role": self.role}
        if self.content is not None:
            result["content"] = self.content.dump(camel_case=camel_case)
        if self.data is not None:
            result["data"] = [item.dump(camel_case=camel_case) for item in self.data]
        if self.reasoning is not None:
            result["reasoning"] = [item.dump(camel_case=camel_case) for item in self.reasoning]
        if self.actions is not None:
            result["actions"] = [item.dump(camel_case=camel_case) for item in self.actions]
        return result

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> AgentMessage:
        content = MessageContent._load(data["content"]) if "content" in data else None
        data_items = [AgentDataItem._load(item, cognite_client) for item in data.get("data", [])]
        reasoning_items = [AgentReasoningItem._load(item, cognite_client) for item in data.get("reasoning", [])]
        action_calls = [ActionCall._load(item, cognite_client) for item in data.get("actions", [])]
        return cls(
            content=content,
            data=data_items if data_items else None,
            reasoning=reasoning_items if reasoning_items else None,
            actions=action_calls if action_calls else None,
            role=data["role"],
        )


class AgentMessageList(CogniteResourceList[AgentMessage]):
    """List of agent messages."""

    _RESOURCE = AgentMessage


class AgentChatResponse(CogniteResource):
    """Response from agent chat.

    Args:
        agent_external_id (str): The external ID of the agent.
        messages (AgentMessageList): The response messages from the agent.
        type (str): The response type.
        cursor (str | None): Cursor for conversation continuation.
    """

    def __init__(
        self,
        agent_external_id: str,
        messages: AgentMessageList,
        type: str,
        cursor: str | None = None,
    ) -> None:
        self.agent_external_id = agent_external_id
        self.cursor = cursor
        self.messages = messages
        self.type = type

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result = {
            "agentExternalId" if camel_case else "agent_external_id": self.agent_external_id,
            "response": {
                "cursor": self.cursor,
                "messages": [msg.dump(camel_case=camel_case) for msg in self.messages] if self.messages else [],
                "type": self.type,
            },
        }
        return result

    @property
    def text(self) -> str | None:
        """Get the text content from the first message with text content."""
        if self.messages:
            for message in self.messages:
                if message.content and isinstance(message.content, TextContent):
                    return message.content.text
        return None

    @property
    def action_calls(self) -> list[ActionCall] | None:
        """Get all action calls from all messages."""
        if not self.messages:
            return None
        return [call for msgs in self.messages if msgs.actions for call in msgs.actions] or None

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> AgentChatResponse:
        response_data = data["response"]
        raw_messages = response_data["messages"]
        messages = [AgentMessage._load(msg, cognite_client) for msg in raw_messages]
        messages_list = AgentMessageList(messages)

        instance = cls(
            agent_external_id=data["agentExternalId"],
            cursor=response_data.get("cursor"),
            messages=messages_list,
            type=response_data["type"],
        )

        return instance



from __future__ import annotations

from abc import ABC
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteFilter,
    CogniteObject,
    CogniteResourceList,
    ExternalIDTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class LabelDefinitionCore(WriteableCogniteResource["LabelDefinitionWrite"], ABC):
    """A label definition is a globally defined label that can later be attached to resources (e.g., assets). For example, can you define a "Pump" label definition and attach that label to your pump assets.
    This is the parent for the reading and writing versions.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        name (str | None): Name of the label.
        description (str | None): Description of the label.
        data_set_id (int | None): The id of the dataset this label belongs to.
    """

    def __init__(
        self,
        external_id: str,
        name: str | None,
        description: str | None,
        data_set_id: int | None,
    ) -> None:
        self.external_id = external_id
        self.name = name
        self.description = description
        self.data_set_id = data_set_id


class LabelDefinition(LabelDefinitionCore):
    """A label definition is a globally defined label that can later be attached to resources (e.g., assets). For example, can you define a "Pump" label definition and attach that label to your pump assets.
    This is the reading version of the LabelDefinition class. It is used when retrieving existing label definitions.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        name (str): Name of the label.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        description (str | None): Description of the label.
        data_set_id (int | None): The id of the dataset this label belongs to.
        cognite_client (AsyncCogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        external_id: str,
        name: str,
        created_time: int,
        description: str | None,
        data_set_id: int | None,
        cognite_client: AsyncCogniteClient | None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            name=name,
            description=description,
            data_set_id=data_set_id,
        )
        self.created_time = created_time
        self._cognite_client = cast("AsyncCogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            name=resource["name"],
            created_time=resource["createdTime"],
            description=resource.get("description"),
            data_set_id=resource.get("dataSetId"),
            cognite_client=cognite_client,
        )

    def as_write(self) -> LabelDefinitionWrite:
        """Returns this LabelDefinition in its writing version."""
        if self.external_id is None or self.name is None:
            raise ValueError("External ID and name are required for the writing version of a label definition.")
        return LabelDefinitionWrite(
            external_id=self.external_id,
            name=self.name,
            description=self.description,
            data_set_id=self.data_set_id,
        )


class LabelDefinitionWrite(LabelDefinitionCore):
    """A label definition is a globally defined label that can later be attached to resources (e.g., assets). For example, can you define a "Pump" label definition and attach that label to your pump assets.
    This is the writing version of the LabelDefinition class. It is used when creating new label definitions.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        name (str): Name of the label.
        description (str | None): Description of the label.
        data_set_id (int | None): The id of the dataset this label belongs to.
    """

    def __init__(
        self,
        external_id: str,
        name: str,
        description: str | None = None,
        data_set_id: int | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            name=name,
            description=description,
            data_set_id=data_set_id,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> LabelDefinitionWrite:
        return cls(
            external_id=resource["externalId"],
            name=resource["name"],
            description=resource.get("description"),
            data_set_id=resource.get("dataSetId"),
        )

    def as_write(self) -> LabelDefinitionWrite:
        """Returns this LabelDefinitionWrite instance."""
        return self


class LabelDefinitionFilter(CogniteFilter):
    """Filter on labels definitions with strict matching.

    Args:
        name (str | None): Returns the label definitions matching that name.
        external_id_prefix (str | None): filter label definitions with external ids starting with the prefix specified
        data_set_ids (list[dict[str, Any]] | None): Only include labels that belong to these datasets.
        cognite_client (AsyncCogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        name: str | None = None,
        external_id_prefix: str | None = None,
        data_set_ids: list[dict[str, Any]] | None = None,
        cognite_client: AsyncCogniteClient | None = None,
    ) -> None:
        self.name = name
        self.external_id_prefix = external_id_prefix
        self.data_set_ids = data_set_ids
        self._cognite_client = cast("AsyncCogniteClient", cognite_client)


class LabelDefinitionWriteList(CogniteResourceList[LabelDefinitionWrite], ExternalIDTransformerMixin):
    _RESOURCE = LabelDefinitionWrite


class LabelDefinitionList(
    WriteableCogniteResourceList[LabelDefinitionWrite, LabelDefinition], ExternalIDTransformerMixin
):
    _RESOURCE = LabelDefinition

    def as_write(self) -> LabelDefinitionWriteList:
        """Returns this LabelDefinitionList in its writing version."""
        return LabelDefinitionWriteList(
            [item.as_write() for item in self.data], cognite_client=self._get_cognite_client()
        )


class Label(CogniteObject):
    """A label assigned to a resource.

    Args:
        external_id (str): The external id to the attached label.
    """

    def __init__(self, external_id: str) -> None:
        self.external_id = external_id

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(external_id=resource["externalId"])

    @classmethod
    def _load_list(
        cls,
        labels: SequenceNotStr[str | dict | LabelDefinitionCore | Label]
        | Sequence[dict | LabelDefinitionCore | Label]
        | None,
    ) -> list[Label] | None:
        def convert_label(label: Label | str | LabelDefinitionCore | dict) -> Label:
            if isinstance(label, Label):
                return label
            elif isinstance(label, str):
                return Label(label)
            elif isinstance(label, LabelDefinitionCore):
                return Label(label.external_id)
            elif isinstance(label, dict):
                if "externalId" in label:
                    return Label(label["externalId"])
                if "external_id" in label:
                    return Label(label["external_id"])
            raise ValueError(f"Could not parse label: {label}")

        if labels is None:
            return None
        return [convert_label(label) for label in labels]


class LabelFilter(CogniteFilter):
    """Return only the resource matching the specified label constraints.

    Args:
        contains_any (list[str] | None): The resource item contains at least one of the listed labels. The labels are defined by a list of external ids.
        contains_all (list[str] | None): The resource item contains all the listed labels. The labels are defined by a list of external ids.
        cognite_client (AsyncCogniteClient | None): The client to associate with this object.

    Examples:

            List only resources marked as a PUMP and VERIFIED::

                >>> from cognite.client.data_classes import LabelFilter
                >>> my_label_filter = LabelFilter(contains_all=["PUMP", "VERIFIED"])

            List only resources marked as a PUMP or as a VALVE::

                >>> from cognite.client.data_classes import LabelFilter
                >>> my_label_filter = LabelFilter(contains_any=["PUMP", "VALVE"])
    """

    def __init__(
        self,
        contains_any: list[str] | None = None,
        contains_all: list[str] | None = None,
        cognite_client: AsyncCogniteClient | None = None,
    ) -> None:
        self.contains_any = contains_any
        self.contains_all = contains_all
        self._cognite_client = cast("AsyncCogniteClient", cognite_client)

    @classmethod
    def _load(cls, label_filter: dict[str, Any]) -> LabelFilter:
        return cls(
            contains_any=(any_ := label_filter.get("containsAny")) and [item["externalId"] for item in any_],
            contains_all=(all_ := label_filter.get("containsAll")) and [item["externalId"] for item in all_],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        dumped: dict[str, Any] = {}
        if self.contains_any:
            dumped["containsAny"] = [
                {"externalId" if camel_case else "external_id": item} for item in self.contains_any
            ]
        if self.contains_all:
            dumped["containsAll"] = [
                {"externalId" if camel_case else "external_id": item} for item in self.contains_all
            ]
        return dumped



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, cast

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
    UnknownCogniteObject,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.data_modeling.ids import ViewId

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient

ColumnType: TypeAlias = Literal[
    "BIGINT",
    "BIGINT[]",
    "BOOL",
    "BOOL[]",
    "BYTEA",
    "DOUBLE PRECISION",
    "DOUBLE PRECISION[]",
    "JSON",
    "JSON[]",
    "REAL",
    "REAL[]",
    "TEXT",
    "TEXT[]",
    "TIMESTAMPTZ",
    "VARCHAR",
    "VARCHAR[]",
]


@dataclass
class RawTableOptions(CogniteObject):
    database: str
    table: str
    primary_key: str | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            database=resource["database"],
            table=resource["table"],
            primary_key=resource.get("primaryKey"),
        )


@dataclass
class Column(CogniteResource):
    name: str
    type: ColumnType

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            name=resource["propertyName"],
            type=resource["type"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "propertyName": self.name,
            "type": self.type,
        }


class ColumnList(CogniteResourceList[Column]):
    _RESOURCE = Column

    @classmethod
    def _load_columns(cls, data: dict[str, Any] | Iterable[dict[str, Any]]) -> ColumnList:
        if not isinstance(data, dict):
            return cls._load(data, None)
        columns = cls([], None)
        for name, column_data in data.items():
            columns.append(
                Column(
                    name=name,
                    type=column_data["type"],
                )
            )
        return columns

    def _dump_columns(self) -> dict[str, Any]:
        return {column.name: {"type": column.type} for column in self.data}


class _TableCore(WriteableCogniteResource["TableWrite"], ABC):
    _type: str

    def __init__(self, tablename: str):
        self.tablename = tablename

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {"tablename": self.tablename, "type": self._type}


class TableWrite(_TableCore, ABC):
    """View and create foreign **tables** for a given **user**.

    This is the write/request format of the table.
    """

    def as_write(self) -> Self:
        return self

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        type_ = resource.get("type")
        if type_ is None and hasattr(cls, "_type"):
            type_ = cls._type
        elif type_ is None:
            raise KeyError("type")
        try:
            return cast(Self, _TABLE_WRITE_CLASS_BY_TYPE[type_]._load_table(resource))
        except KeyError:
            raise TypeError(f"Unknown source type: {type_}")

    @classmethod
    @abstractmethod
    def _load_table(cls, data: dict[str, Any]) -> Self:
        raise NotImplementedError()


class RawTableWrite(TableWrite):
    """Foreign tables.

    This is the read/response format of the raw table.

    Args:
        tablename (str): Name of the foreign table.
        options (RawTableOptions): Table options
        columns (Sequence[Column] | ColumnList): Foreign table columns.

    """

    _type = "raw_rows"

    def __init__(self, tablename: str, options: RawTableOptions, columns: Sequence[Column] | ColumnList) -> None:
        super().__init__(tablename=tablename)
        self.options = options
        self.columns: ColumnList = ColumnList(columns)

    @classmethod
    def _load_table(cls, data: dict[str, Any]) -> Self:
        return cls(
            tablename=data["tablename"],
            options=RawTableOptions._load(data["options"]),
            columns=ColumnList._load_columns(data["columns"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        output["options"] = self.options.dump(camel_case=camel_case)
        output["columns"] = self.columns._dump_columns()
        return output


class ViewTableWrite(TableWrite):
    """Foreign tables.

    This is the read/response format of the custom table.

    Args:
        tablename (str): Name of the foreign table.
        options (ViewId): Table options
    """

    _type = "view"

    def __init__(self, tablename: str, options: ViewId) -> None:
        super().__init__(tablename=tablename)
        self.options = options

    @classmethod
    def _load_table(cls, data: dict[str, Any]) -> Self:
        return cls(
            tablename=data["tablename"],
            options=ViewId.load(data["options"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        output["options"] = self.options.dump(camel_case=camel_case, include_type=False)
        return output


class Table(_TableCore, ABC):
    """Foreign tables.

    This is the read/response format of the custom table.

    Args:
        tablename (str): Name of the foreign table.
        created_time (int | None): Time when the table was created

    """

    def __init__(self, tablename: str, created_time: int | None = None) -> None:
        super().__init__(tablename=tablename)
        self.created_time = created_time

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        type_ = resource.get("type")
        if type_ is None and hasattr(cls, "_type"):
            type_ = cls._type
        elif type_ is None:
            raise KeyError("type")
        table_cls = _TABLE_CLASS_BY_TYPE.get(type_)
        if table_cls is None:
            return UnknownCogniteObject(resource)  # type: ignore[return-value]
        return cast(Self, table_cls._load_table(resource))

    @classmethod
    @abstractmethod
    def _load_table(cls, data: dict[str, Any]) -> Self:
        raise NotImplementedError


class RawTable(Table):
    """Foreign tables.

    This is the read/response format of the raw table.

    Args:
        tablename (str): Name of the foreign table.
        options (RawTableOptions): Table options
        columns (ColumnList): Foreign table columns.
        created_time (int | None): Time when the table was created.

    """

    _type = "raw_rows"

    def __init__(
        self, tablename: str, options: RawTableOptions, columns: ColumnList, created_time: int | None = None
    ) -> None:
        super().__init__(tablename=tablename, created_time=created_time)
        self.options = options
        self.columns = columns

    @classmethod
    def _load_table(cls, data: dict[str, Any]) -> Self:
        return cls(
            tablename=data["tablename"],
            options=RawTableOptions._load(data["options"]),
            columns=ColumnList._load_columns(data["columns"]),
            created_time=data.get("createdTime"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        output["options"] = self.options.dump(camel_case=camel_case)
        output["columns"] = self.columns.dump(camel_case=camel_case)
        return output

    def as_write(self) -> RawTableWrite:
        return RawTableWrite(
            tablename=self.tablename,
            options=self.options,
            columns=self.columns,
        )


class ViewTable(Table):
    """Foreign tables.

    This is the read/response format of the custom table.

    Args:
        tablename (str): Name of the foreign table.
        options (ViewId): Table options
        created_time (int | None): Time when the table was created.
    """

    _type = "view"

    def __init__(self, tablename: str, options: ViewId, created_time: int | None = None) -> None:
        super().__init__(tablename=tablename, created_time=created_time)
        self.options = options

    @classmethod
    def _load_table(cls, data: dict[str, Any]) -> Self:
        return cls(
            tablename=data["tablename"],
            options=ViewId.load(data["options"]),
            created_time=data.get("created_time"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        output["options"] = self.options.dump(camel_case=camel_case)
        return output

    def as_write(self) -> ViewTableWrite:
        return ViewTableWrite(
            tablename=self.tablename,
            options=self.options,
        )


class TableWriteList(CogniteResourceList[TableWrite]):
    _RESOURCE = TableWrite


class TableList(WriteableCogniteResourceList[TableWrite, Table]):
    _RESOURCE = Table

    def as_write(self) -> TableWriteList:
        return TableWriteList([item.as_write() for item in self.data])


_TABLE_WRITE_CLASS_BY_TYPE: dict[str, type[TableWrite]] = {
    subclass._type: subclass  # type: ignore[type-abstract]
    for subclass in TableWrite.__subclasses__()
}

_TABLE_CLASS_BY_TYPE: dict[str, type[Table]] = {
    subclass._type: subclass  # type: ignore[type-abstract]
    for subclass in Table.__subclasses__()
}



from __future__ import annotations

from collections.abc import Sequence
from math import ceil
from typing import TYPE_CHECKING, Any, Literal, TypeVar, overload

from cognite.client._api_client import APIClient
from cognite.client.data_classes._base import CogniteResource
from cognite.client.data_classes.contextualization import (
    DetectJobBundle,
    DiagramConvertResults,
    DiagramDetectConfig,
    DiagramDetectResults,
    FileReference,
)
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils._experimental import FeaturePreviewWarning
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient
    from cognite.client.config import ClientConfig

_T = TypeVar("_T")


class DiagramsAPI(APIClient):
    _RESOURCE_PATH = "/context/diagram"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: AsyncCogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        # https://developer.cognite.com/api#tag/Engineering-diagrams/operation/diagramDetect
        self._DETECT_API_FILE_LIMIT = 50
        self._DETECT_API_STATUS_JOB_LIMIT = 1000
        self._detect_beta_params_warning = FeaturePreviewWarning(
            api_maturity="beta",
            sdk_maturity="beta",
            feature_name="Support for diagram detect 'configuration' and 'pattern_mode' parameters",
        )

    @staticmethod
    def _list_from_instance_or_list(
        instance_or_list: Sequence[_T] | _T | None, instance_type: type[_T], error_message: str
    ) -> Sequence[_T]:
        if instance_or_list is None:
            return []
        if isinstance(instance_or_list, instance_type):
            return [instance_or_list]
        if isinstance(instance_or_list, list) and all(isinstance(x, instance_type) for x in instance_or_list):
            return instance_or_list
        raise TypeError(error_message)

    @classmethod
    def _process_file_ids(
        cls,
        ids: Sequence[int] | int | None,
        external_ids: SequenceNotStr[str] | str | None,
        instance_ids: Sequence[NodeId] | NodeId | None,
        file_references: Sequence[FileReference] | FileReference | None,
    ) -> list[dict]:
        id_list = cls._list_from_instance_or_list(ids, int, "ids must be int or list of int")
        extid_list = cls._list_from_instance_or_list(external_ids, str, "external_ids must be str or list of str")
        instance_id_list = cls._list_from_instance_or_list(
            instance_ids, NodeId, "instance_ids must be NodeId or list of NodeId"
        )
        file_references = cls._list_from_instance_or_list(
            file_references, FileReference, "file_references must be FileReference or list of FileReference"
        )
        # Handle empty lists
        if not (extid_list or id_list or instance_id_list or file_references):
            raise ValueError("No ids, external ids or file references specified")

        id_objs = [{"fileId": id} for id in id_list]
        external_id_objs = [{"fileExternalId": external_id} for external_id in extid_list]
        instance_id_objs = [
            {"fileInstanceId": instance_id.dump(camel_case=True, include_instance_type=False)}
            for instance_id in instance_id_list
        ]
        file_reference_objects = [file_reference.to_api_item() for file_reference in file_references]
        processed_ids: list[dict] = [*id_objs, *external_id_objs, *instance_id_objs, *file_reference_objects]
        return processed_ids

    @overload
    async def detect(
        self,
        entities: Sequence[dict | CogniteResource],
        search_field: str = "name",
        partial_match: bool = False,
        min_tokens: int = 2,
        file_ids: int | Sequence[int] | None = None,
        file_external_ids: str | SequenceNotStr[str] | None = None,
        file_instance_ids: NodeId | Sequence[NodeId] | None = None,
        file_references: list[FileReference] | FileReference | None = None,
        pattern_mode: bool = False,
        configuration: DiagramDetectConfig | None = None,
        *,
        multiple_jobs: Literal[False] = False,
    ) -> DiagramDetectResults: ...

    @overload
    async def detect(
        self,
        entities: Sequence[dict | CogniteResource],
        search_field: str = "name",
        partial_match: bool = False,
        min_tokens: int = 2,
        file_ids: int | Sequence[int] | None = None,
        file_external_ids: str | SequenceNotStr[str] | None = None,
        file_instance_ids: NodeId | Sequence[NodeId] | None = None,
        file_references: list[FileReference] | FileReference | None = None,
        pattern_mode: bool = False,
        configuration: DiagramDetectConfig | None = None,
        *,
        multiple_jobs: Literal[True],
    ) -> tuple[DetectJobBundle, list[dict[str, Any]]]: ...

    async def detect(
        self,
        entities: Sequence[dict | CogniteResource],
        search_field: str = "name",
        partial_match: bool = False,
        min_tokens: int = 2,
        file_ids: int | Sequence[int] | None = None,
        file_external_ids: str | SequenceNotStr[str] | None = None,
        file_instance_ids: NodeId | Sequence[NodeId] | None = None,
        file_references: list[FileReference] | FileReference | None = None,
        pattern_mode: bool | None = None,
        configuration: DiagramDetectConfig | None = None,
        *,
        multiple_jobs: bool = False,
    ) -> DiagramDetectResults | tuple[DetectJobBundle, list[dict[str, Any]]]:
        """`Detect annotations in engineering diagrams <https://developer.cognite.com/api#tag/Engineering-diagrams/operation/diagramDetect>`_

        Note:
            All users on this CDF subscription with assets read-all and files read-all capabilities in the project,
            are able to access the data sent to this endpoint.

        Args:
            entities (Sequence[dict | CogniteResource]): List of entities to detect
            search_field (str): If entities is a list of dictionaries, this is the key to the values to detect in the PnId
            partial_match (bool): Allow for a partial match (e.g. missing prefix).
            min_tokens (int): Minimal number of tokens a match must be based on
            file_ids (int | Sequence[int] | None): ID of the files, should already be uploaded in the same tenant.
            file_external_ids (str | SequenceNotStr[str] | None): File external ids, alternative to file_ids and file_references.
            file_instance_ids (NodeId | Sequence[NodeId] | None): Files to detect in, specified by instance id.
            file_references (list[FileReference] | FileReference | None): File references (id, external_id or instance_id), and first_page and last_page to specify page ranges per file. Each reference can specify up to 50 pages. Providing a page range will also make the page count of the document a part of the response.
            pattern_mode (bool | None): If True, entities must be provided with a sample field. This enables detecting tags that are similar to the sample, but not necessarily identical. Defaults to None.
            configuration (DiagramDetectConfig | None): Additional configuration for the detect algorithm. See `DiagramDetectConfig` class documentation and `beta API docs <https://api-docs.cognite.com/20230101-beta/tag/Engineering-diagrams/operation/diagramDetect/#!path=configuration&t=request>`_.
            multiple_jobs (bool): Enables you to publish multiple jobs. If True the method returns a tuple of DetectJobBundle and list of potentially unposted files. If False it will return a single DiagramDetectResults. Defaults to False.
        Returns:
            DiagramDetectResults | tuple[DetectJobBundle, list[dict[str, Any]]]: Resulting queued job or a bundle of jobs and a list of unposted files. Note that the .result property of the job or job bundle will block waiting for results.

        Note:
            The results are not written to CDF, to create annotations based on detected entities use `AnnotationsAPI`.

        Examples:
                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.contextualization import FileReference
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> detect_job = client.diagrams.detect(
                ...     entities=[
                ...         {"userDefinedField": "21PT1017","ignoredField": "AA11"},
                ...         {"userDefinedField": "21PT1018"}],
                ...     search_field="userDefinedField",
                ...     partial_match=True,
                ...     min_tokens=2,
                ...     file_ids=[101],
                ...     file_external_ids=["Test1"],
                ...     file_references=[
                ...         FileReference(id=20, first_page=1, last_page=10),
                ...         FileReference(external_id="ext_20", first_page=11, last_page=20)
                ...     ])
                >>> result = detect_job.get_result()
                >>> print(result)
                <code>
                {
                    'items': [
                        {'fileId': 101, 'annotations': []},
                        {'fileExternalId': 'Test1', 'fileId: 1, 'annotations': []},
                        {'fileId': 20, 'fileExternalId': 'ext_20', 'annotations': [], 'pageCount': 17},
                        {
                            'fileId': 20,
                            'fileExternalId': 'ext_20',
                            'annotations': [
                                {
                                    'text': '21PT1017',
                                    'entities': [{"userDefinedField": "21PT1017","ignoredField": "AA11"}],
                                    'region': {
                                        'page': 12,
                                        'shape': 'rectangle',
                                        'vertices': [
                                            {'x': 0.01, 'y': 0.01},
                                            {'x': 0.01, 'y': 0.02},
                                            {'x': 0.02, 'y': 0.02},
                                            {'x': 0.02, 'y': 0.01}
                                        ]
                                    }
                                }
                            ],
                            'pageCount': 17
                        }
                    ]
                }
                </code>

            To use beta configuration options you can use a dictionary or `DiagramDetectConfig` object for convenience:

                >>> from cognite.client.data_classes.contextualization import ConnectionFlags, DiagramDetectConfig
                >>> config = DiagramDetectConfig(
                ...     remove_leading_zeros=True,
                ...     connection_flags=ConnectionFlags(
                ...         no_text_inbetween=True,
                ...         natural_reading_order=True,
                ...     )
                ... )
                >>> job = client.diagrams.detect(entities=[{"name": "A1"}], file_id=123, config=config)

            Check the documentation for `DiagramDetectConfig` for more information on the available options.
        """
        items = self._process_file_ids(file_ids, file_external_ids, file_instance_ids, file_references)
        entities_dumped = [
            entity.dump(camel_case=True) if isinstance(entity, CogniteResource) else entity for entity in entities
        ]

        if not multiple_jobs:
            return await self.__run_detect_job(
                items=items,
                entities=entities_dumped,
                partial_match=partial_match,
                search_field=search_field,
                min_tokens=min_tokens,
                pattern_mode=pattern_mode,
                configuration=configuration,
            )

        num_new_jobs = ceil(len(items) / self._DETECT_API_FILE_LIMIT)
        if num_new_jobs > self._DETECT_API_STATUS_JOB_LIMIT:
            raise ValueError(
                f"Number of jobs exceed limit of: '{self._DETECT_API_STATUS_JOB_LIMIT}'. Number of jobs: '{num_new_jobs}'"
            )

        jobs: list[DiagramDetectResults] = []
        unposted_files: list[dict[str, Any]] = []
        for i in range(num_new_jobs):
            batch = items[(self._DETECT_API_FILE_LIMIT * i) : self._DETECT_API_FILE_LIMIT * (i + 1)]
            try:
                job = await self.__run_detect_job(
                    items=items,
                    entities=entities_dumped,
                    partial_match=partial_match,
                    search_field=search_field,
                    min_tokens=min_tokens,
                    pattern_mode=pattern_mode,
                    configuration=configuration,
                )
                jobs.append(job)
            except CogniteAPIError as exc:
                unposted_files.append({"error": str(exc), "files": batch})

        res = DetectJobBundle(cognite_client=self._cognite_client, job_ids=[j.job_id for j in jobs])
        return res, unposted_files

    async def __run_detect_job(
        self,
        items: list[dict],
        entities: list[dict],
        partial_match: bool,
        search_field: str,
        min_tokens: int,
        pattern_mode: bool | None,
        configuration: DiagramDetectConfig | None,
    ) -> DiagramDetectResults:
        api_subversion = None
        beta_parameters: dict = {}
        if pattern_mode is not None or configuration is not None:
            if pattern_mode is not None:
                beta_parameters["patternMode"] = pattern_mode
            if configuration is not None:
                beta_parameters["configuration"] = configuration.dump()
            self._detect_beta_params_warning.warn()
            if self._api_subversion and not self._api_subversion.endswith("beta"):
                api_subversion = f"{self._api_subversion}-beta"

        body = {
            "items": items,
            "entities": entities,
            "partialMatch": partial_match,
            "searchField": search_field,
            "minTokens": min_tokens,
            **beta_parameters,
        }

        response = await self._post(
            f"{self._RESOURCE_PATH}/detect",
            json=body,
            api_subversion=api_subversion,
        )
        return DiagramDetectResults._load_with_job_token(
            data=response.json(),
            headers=response.headers,
            cognite_client=self._cognite_client,
        )

    async def get_detect_jobs(self, job_ids: list[int]) -> list[DiagramDetectResults]:
        res = await self._cognite_client.diagrams._post("/context/diagram/detect/status", json={"items": job_ids})
        jobs = res.json()["items"]
        return [
            DiagramDetectResults._load_with_job_token(job, headers=res.headers, cognite_client=self._cognite_client)
            for job in jobs
        ]

    async def convert(self, detect_job: DiagramDetectResults) -> DiagramConvertResults:
        """Convert a P&ID to interactive SVGs where the provided annotations are highlighted.

        Note:
            Will automatically wait for the detect job to complete before starting the conversion.

        Args:
            detect_job (DiagramDetectResults): detect job

        Returns:
            DiagramConvertResults: Resulting queued job.

        Examples:

            Run a detection job, then convert the results:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> detect_job = client.diagrams.detect(...)
                >>> client.diagrams.convert(detect_job=detect_job)

        """
        result = await detect_job.get_result_async()
        if any(item.get("page_range") is not None for item in result["items"]):
            raise NotImplementedError("Can not run convert on a detect job that used the page range feature")
        items = [{"annotations": item.get("annotations"), "fileId": item.get("fileId")} for item in result["items"]]
        response = await self._post(f"{self._RESOURCE_PATH}/convert", json={"items": items})
        return DiagramConvertResults._load_with_job_token(
            data=response.json(),
            headers=response.headers,
            cognite_client=self._cognite_client,
        )



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any
from urllib.parse import quote

if TYPE_CHECKING:
    from cognite.client._basic_api_client import BasicAsyncAPIClient


NON_RETRYABLE_CREATE_DELETE_RESOURCE_PATHS: tuple[str, ...] = (
    "annotations",
    "assets",
    "context/entitymatching",
    "datasets",
    "documents",
    "events",
    "extpipes",
    "extpipes/config",
    "extpipes/runs",
    "files",
    "functions",
    "functions/[^/]+/call",
    "functions/schedules",
    "geospatial",
    "geospatial/crs",
    "geospatial/featuretypes",
    "geospatial/featuretypes/[^/]+/features",
    "hostedextractors",
    "labels",
    "postgresgateway",
    "profiles",
    "raw/dbs$",
    "raw/dbs/[^/]+/tables$",
    "relationships",
    "sequences",
    "simulators",
    "simulators/models",
    "simulators/models/revisions",
    "simulators/models/routines",
    "simulators/models/routines/revisions",
    "timeseries",
    "transformations",
    "transformations/schedules",
    "3d/models",
    "3d/models/[^/]+/revisions",
    "3d/models/[^/]+/revisions/[^/]+/mappings",
    "3d/models/[^/]+/revisions/[^/]+/nodes",
)
NON_IDEMPOTENT_POST_ENDPOINT_REGEX_PATTERN: re.Pattern[str] = re.compile(
    r"|".join(
        rf"^/{path}(\?.*)?$"
        for path in (
            f"({r'|'.join(NON_RETRYABLE_CREATE_DELETE_RESOURCE_PATHS)})(/delete)?$",
            "ai/tools/documents/task",
            "annotations/suggest",
            "extpipes/config/revert",
            "transformations/cancel",
            "transformations/notifications",
            "transformations/run",
        )
    )
)
VALID_URL_PATTERN = re.compile(r"^https?://[a-z\d.:\-]+(?:/api/v1/projects/[^/]+)?((/[^\?]+)?(\?.+)?)")
VALID_METHODS = {"GET", "POST", "PUT", "DELETE", "PATCH"}


def resolve_url(api_client: BasicAsyncAPIClient, method: str, url_path: str) -> tuple[bool, str]:
    if not url_path.startswith("/"):
        raise ValueError("URL path must start with '/'")

    full_url = api_client._base_url_with_base_path + url_path
    is_retryable = validate_url_and_return_retryability(method, full_url)
    return is_retryable, full_url


def validate_url_and_return_retryability(method: str, full_url: str) -> bool:
    if method not in VALID_METHODS:
        raise ValueError(f"Method {method} is not valid. Must be one of {VALID_METHODS}")

    if valid_url := VALID_URL_PATTERN.match(full_url):
        is_retryable = can_be_retried(method, valid_url.group(1))
        return is_retryable

    raise ValueError(f"URL {full_url} is not valid. Cannot resolve whether or not it is retryable")


def can_be_retried(method: str, path: str) -> bool:
    match method:
        case "GET" | "PUT" | "PATCH":
            return True
        case "POST" if not NON_IDEMPOTENT_POST_ENDPOINT_REGEX_PATTERN.match(path):
            return True
        case _:
            return False


def interpolate_and_url_encode(path: str, *args: Any) -> str:
    return path.format(*(quote(str(arg), safe="") for arg in args))



"""
===============================================================================
c23bdbc81be7b11616bd3dae81e7ef44
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from cognite.client import AsyncCogniteClient
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.utils._async_helpers import run_sync


class Sync3DFilesAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    def retrieve(self, id: int) -> bytes:
        """
        `Retrieve the contents of a 3d file by id. <https://developer.cognite.com/api#tag/3D-Files/operation/get3DFile>`_

        Args:
            id (int): The id of the file to retrieve.

        Returns:
            bytes: The contents of the file.

        Example:

            Retrieve the contents of a 3d file by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.three_d.files.retrieve(1)
        """
        return run_sync(self.__async_client.three_d.files.retrieve(id=id))



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import asyncio
    from datetime import datetime


@dataclass
class SubscriptionContext:
    """Manages an active subscription to an instance sync query."""

    last_successful_sync: datetime | None = None
    last_successful_callback: datetime | None = None
    _task: asyncio.Task | None = None

    def cancel(self) -> None:
        if self._task and not self._task.done():
            self._task.cancel()

    def is_running(self) -> bool:
        return self._task is not None and not self._task.done()

    def is_alive(self) -> bool:
        return self.is_running()



from __future__ import annotations

import math
import operator as op
from collections import defaultdict
from collections.abc import Callable, Iterator, Sequence
from itertools import chain
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, cast

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer

from cognite.client._constants import NUMPY_IS_AVAILABLE
from cognite.client._proto.data_point_list_response_pb2 import DataPointListItem
from cognite.client._proto.data_points_pb2 import (
    AggregateDatapoint,
    NumericDatapoint,
    StringDatapoint,
)
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.data_classes.datapoints import (
    DatapointsQuery,
    MaxDatapoint,
    MaxDatapointWithStatus,
    MaxOrMinDatapoint,
    MinDatapoint,
    MinDatapointWithStatus,
)
from cognite.client.utils.useful_types import SequenceNotStr

if NUMPY_IS_AVAILABLE:
    import numpy as np

if TYPE_CHECKING:
    import numpy.typing as npt

AggregateDatapoints = RepeatedCompositeFieldContainer[AggregateDatapoint]
NumericDatapoints = RepeatedCompositeFieldContainer[NumericDatapoint]
StringDatapoints = RepeatedCompositeFieldContainer[StringDatapoint]

DatapointAny = AggregateDatapoint | NumericDatapoint | StringDatapoint
DatapointsAny = AggregateDatapoints | NumericDatapoints | StringDatapoints

DatapointRaw = NumericDatapoint | StringDatapoint
DatapointsRaw = NumericDatapoints | StringDatapoints

RawDatapointValue = float | str
DatapointsId = int | DatapointsQuery | Sequence[int | DatapointsQuery]
DatapointsExternalId = str | DatapointsQuery | SequenceNotStr[str | DatapointsQuery]
DatapointsInstanceId = NodeId | DatapointsQuery | Sequence[NodeId | DatapointsQuery]


class DpsUnpackFns:
    ts: Callable[[DatapointAny], int] = op.attrgetter("timestamp")
    raw_dp: Callable[[DatapointRaw], RawDatapointValue] = op.attrgetter("value")

    @staticmethod
    def custom_from_aggregates(lst: list[str]) -> Callable[[AggregateDatapoint], tuple[float, ...]]:
        return op.attrgetter(*lst)

    # Status is a nested object in the response (code+symbol). Since most dps is expected to be good
    # (code 0), status is not returned for these:
    status_code: Callable[[DatapointRaw], int] = op.attrgetter("status.code")  # Gives 0 by default when missing

    @staticmethod
    def status_symbol(dp: DatapointRaw) -> str:
        return dp.status.symbol or "Good"  # Gives empty str when missing, so we set 'Good' manually

    # When datapoints with bad status codes are not ignored, value may be missing:
    @staticmethod
    def nullable_raw_dp(dp: DatapointRaw) -> float | str:
        # We pretend like float is always returned to not break every dps annot. in the entire SDK..
        return dp.value if not dp.nullValue else None  # type: ignore [return-value]

    # minDatapoint and maxDatapoint are also objects in the response. The proto lookups doesn't fail,
    # so we must be very careful to only attach status codes if requested.
    @staticmethod
    def min_datapoint(agg_dp: AggregateDatapoint) -> MinDatapoint:
        dp = agg_dp.minDatapoint
        return MinDatapoint(dp.timestamp, dp.value)

    @staticmethod
    def max_datapoint(agg_dp: AggregateDatapoint) -> MaxDatapoint:
        dp = agg_dp.maxDatapoint
        return MaxDatapoint(dp.timestamp, dp.value)

    @staticmethod
    def min_datapoint_with_status(agg_dp: AggregateDatapoint) -> MinDatapointWithStatus:
        dp = agg_dp.minDatapoint
        return MinDatapointWithStatus(
            dp.timestamp, dp.value, DpsUnpackFns.status_code(dp), DpsUnpackFns.status_symbol(dp)
        )

    @staticmethod
    def max_datapoint_with_status(agg_dp: AggregateDatapoint) -> MaxDatapointWithStatus:
        dp = agg_dp.maxDatapoint
        return MaxDatapointWithStatus(
            dp.timestamp, dp.value, DpsUnpackFns.status_code(dp), DpsUnpackFns.status_symbol(dp)
        )

    # --------------- #
    # Above are functions that operate on single elements
    # Below are functions that operate on containers
    # --------------- #
    @staticmethod
    def extract_timestamps(dps: DatapointsAny) -> list[int]:
        return list(map(DpsUnpackFns.ts, dps))

    @staticmethod
    def extract_timestamps_numpy(dps: DatapointsAny) -> npt.NDArray[np.int64]:
        return np.fromiter(map(DpsUnpackFns.ts, dps), dtype=np.int64, count=len(dps))

    @staticmethod
    def extract_raw_dps(dps: DatapointsRaw) -> list[float | str]:  # Actually: exclusively either one
        return list(map(DpsUnpackFns.raw_dp, dps))

    @staticmethod
    def extract_raw_dps_numpy(dps: DatapointsRaw, dtype: type[np.float64] | type[np.object_]) -> npt.NDArray[Any]:
        return np.fromiter(map(DpsUnpackFns.raw_dp, dps), dtype=dtype, count=len(dps))

    @staticmethod
    def extract_nullable_raw_dps(dps: DatapointsRaw) -> list[float | str]:  # actually list of [... | None]
        return list(map(DpsUnpackFns.nullable_raw_dp, dps))

    @staticmethod
    def extract_nullable_raw_dps_numpy(
        dps: DatapointsRaw, dtype: type[np.float64] | type[np.object_]
    ) -> tuple[npt.NDArray[Any], list[int]]:
        # This is a very hot loop, thus we make some ugly optimizations:
        values = [None] * len(dps)
        missing: list[int] = []
        add_missing = missing.append
        for i, dp in enumerate(map(DpsUnpackFns.nullable_raw_dp, dps)):
            # we use list because of its significantly lower overhead than numpy on single element access:
            values[i] = dp  # type: ignore [call-overload]
            if dp is None:
                add_missing(i)
        arr = np.array(values, dtype=dtype)
        return arr, missing

    @staticmethod
    def extract_status_code(dps: DatapointsRaw) -> list[int]:
        return list(map(DpsUnpackFns.status_code, dps))

    @staticmethod
    def extract_status_code_numpy(dps: DatapointsRaw) -> npt.NDArray[np.uint32]:
        return np.fromiter(map(DpsUnpackFns.status_code, dps), dtype=np.uint32, count=len(dps))

    @staticmethod
    def extract_status_symbol(dps: DatapointsRaw) -> list[str]:
        return list(map(DpsUnpackFns.status_symbol, dps))

    @staticmethod
    def extract_status_symbol_numpy(dps: DatapointsRaw) -> npt.NDArray[np.object_]:
        return np.fromiter(map(DpsUnpackFns.status_symbol, dps), dtype=np.object_, count=len(dps))

    @staticmethod
    def extract_aggregates(
        dps: AggregateDatapoints,
        aggregates: list[str],
        unpack_fn: Callable[[AggregateDatapoint], tuple[float, ...]],
    ) -> list:
        try:
            # Fast method uses multi-key unpacking:
            return list(map(unpack_fn, dps))
        except AttributeError:
            # An aggregate is missing, fallback to slower `getattr`:
            if len(aggregates) == 1:
                return [getattr(dp, aggregates[0], None) for dp in dps]
            else:
                return [tuple(getattr(dp, agg, None) for agg in aggregates) for dp in dps]

    @staticmethod
    def extract_numeric_aggregates_numpy(
        dps: AggregateDatapoints,
        aggregates: list[str],
        unpack_fn: Callable[[AggregateDatapoint], tuple[float, ...]],
        dtype: np.dtype[Any],
    ) -> npt.NDArray[np.float64]:
        try:
            # Fast method uses multi-key unpacking:
            return np.fromiter(map(unpack_fn, dps), dtype=dtype, count=len(dps))
        except AttributeError:
            # An aggregate is missing, fallback to slower `getattr`:
            return np.array([tuple(getattr(dp, agg, math.nan) for agg in aggregates) for dp in dps], dtype=np.float64)

    @staticmethod
    def extract_fn_min_or_max_dp(
        aggregate: Literal["minDatapoint", "maxDatapoint"], include_status: bool
    ) -> Callable[[AggregateDatapoint], MaxOrMinDatapoint]:
        match aggregate, include_status:
            case "minDatapoint", False:
                return DpsUnpackFns.min_datapoint
            case "maxDatapoint", False:
                return DpsUnpackFns.max_datapoint
            case "minDatapoint", True:
                return DpsUnpackFns.min_datapoint_with_status
            case "maxDatapoint", True:
                return DpsUnpackFns.max_datapoint_with_status
            case _:
                raise ValueError(f"Unsupported {aggregate=} and/or {include_status=}")


def ensure_int(val: float, change_nan_to: int = 0) -> int:
    if math.isnan(val):
        return change_nan_to
    return int(val)


def ensure_int_numpy(arr: npt.NDArray[np.float64]) -> npt.NDArray[np.int64]:
    return np.nan_to_num(arr, copy=False, nan=0.0, posinf=np.inf, neginf=-np.inf).astype(np.int64)


def decide_numpy_dtype_from_is_string(is_string: bool) -> type:
    return np.object_ if is_string else np.float64


def get_datapoints_from_proto(res: DataPointListItem) -> DatapointsAny:
    if (dp_type := res.WhichOneof("datapointType")) is not None:
        return getattr(res, dp_type).datapoints
    return cast(DatapointsAny, [])


def get_ts_info_from_proto(res: DataPointListItem) -> dict[str, int | str | bool | NodeId | None]:
    # Note: When 'unit_external_id' is returned, regular 'unit' is ditched
    if res.instanceId and res.instanceId.space:  # res.instanceId evaluates to True even when empty :eyes:
        instance_id = NodeId(res.instanceId.space, res.instanceId.externalId)
    else:
        instance_id = None
    return {
        "id": res.id,
        "external_id": res.externalId,
        "is_string": res.isString,
        "is_step": res.isStep,
        "unit": res.unit,
        "unit_external_id": res.unitExternalId,
        "instance_id": instance_id,
    }


_DataContainer: TypeAlias = defaultdict[tuple[float, ...], list]


def datapoints_in_order(container: _DataContainer) -> Iterator[list]:
    return chain.from_iterable(container[k] for k in sorted(container))


def create_array_from_dps_container(container: _DataContainer) -> npt.NDArray:
    return np.hstack(list(datapoints_in_order(container)))


def create_object_array_from_container(container: _DataContainer) -> npt.NDArray[np.object_]:
    return np.array(create_list_from_dps_container(container), dtype=np.object_)


def create_aggregates_arrays_from_dps_container(container: _DataContainer, n_aggs: int) -> list[npt.NDArray]:
    all_aggs_arr = np.vstack(list(datapoints_in_order(container)))
    return list(map(np.ravel, np.hsplit(all_aggs_arr, n_aggs)))


def create_list_from_dps_container(container: _DataContainer) -> list:
    return list(chain.from_iterable(datapoints_in_order(container)))


def create_aggregates_list_from_dps_container(container: _DataContainer) -> Iterator[list[list]]:
    concatenated = chain.from_iterable(datapoints_in_order(container))
    return map(list, zip(*concatenated))  # rows to columns



"""
===============================================================================
d12a2c456dd92ebcfbac60d16f5ea51b
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from collections.abc import Iterator
from typing import overload

from cognite.client import AsyncCogniteClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes import raw
from cognite.client.utils._async_helpers import SyncIterator, run_sync
from cognite.client.utils.useful_types import SequenceNotStr


class SyncRawTablesAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    @overload
    def __call__(self, db_name: str, chunk_size: None = None) -> Iterator[raw.Table]: ...

    @overload
    def __call__(self, db_name: str, chunk_size: int) -> Iterator[raw.TableList]: ...

    def __call__(
        self, db_name: str, chunk_size: int | None = None, limit: int | None = None
    ) -> Iterator[raw.Table | raw.TableList]:
        """
        Iterate over tables

        Fetches tables as they are iterated over, so you keep a limited number of tables in memory.

        Args:
            db_name (str): Name of the database to iterate over tables for
            chunk_size (int | None): Number of tables to return in each chunk. Defaults to yielding one table a time.
            limit (int | None): Maximum number of tables to return. Defaults to return all items.

        Yields:
            raw.Table | raw.TableList: The tables in the database.
        """
        yield from SyncIterator(self.__async_client.raw.tables(db_name=db_name, chunk_size=chunk_size, limit=limit))  # type: ignore [call-overload]

    @overload
    def create(self, db_name: str, name: str) -> raw.Table: ...

    @overload
    def create(self, db_name: str, name: list[str]) -> raw.TableList: ...

    def create(self, db_name: str, name: str | list[str]) -> raw.Table | raw.TableList:
        """
        `Create one or more tables. <https://developer.cognite.com/api#tag/Raw/operation/createTables>`_

        Args:
            db_name (str): Database to create the tables in.
            name (str | list[str]): A table name or list of table names to create.

        Returns:
            raw.Table | raw.TableList: raw.Table or list of tables that has been created.

        Examples:

            Create a new table in a database:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.raw.tables.create("db1", "table1")
        """
        return run_sync(self.__async_client.raw.tables.create(db_name=db_name, name=name))

    def delete(self, db_name: str, name: str | SequenceNotStr[str]) -> None:
        """
        `Delete one or more tables. <https://developer.cognite.com/api#tag/Raw/operation/deleteTables>`_

        Args:
            db_name (str): Database to delete tables from.
            name (str | SequenceNotStr[str]): A table name or list of table names to delete.

        Examples:

            Delete a list of tables:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.raw.tables.delete("db1", ["table1", "table2"])
        """
        return run_sync(self.__async_client.raw.tables.delete(db_name=db_name, name=name))

    def list(self, db_name: str, limit: int | None = DEFAULT_LIMIT_READ) -> raw.TableList:
        """
        `List tables <https://developer.cognite.com/api#tag/Raw/operation/getTables>`_

        Args:
            db_name (str): The database to list tables from.
            limit (int | None): Maximum number of tables to return. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            raw.TableList: List of requested tables.

        Examples:

            List the first 5 tables:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> table_list = client.raw.tables.list("db1", limit=5)

            Iterate over tables, one-by-one:

                >>> for table in client.raw.tables(db_name="db1"):
                ...     table  # do something with the table

            Iterate over chunks of tables to reduce memory load:

                >>> for table_list in client.raw.tables(db_name="db1", chunk_size=25):
                ...     table_list # do something with the tables
        """
        return run_sync(self.__async_client.raw.tables.list(db_name=db_name, limit=limit))



"""
===============================================================================
341c216ca7805041eb8b81051219edd4
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, overload

from cognite.client import AsyncCogniteClient
from cognite.client._sync_api.unit_system import SyncUnitSystemAPI
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.units import Unit, UnitList
from cognite.client.utils._async_helpers import run_sync
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class SyncUnitAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client
        self.systems = SyncUnitSystemAPI(async_client)

    @overload
    def retrieve(self, external_id: str, ignore_unknown_ids: bool = False) -> None | Unit: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], ignore_unknown_ids: bool = False) -> UnitList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], ignore_unknown_ids: bool = False
    ) -> Unit | UnitList | None:
        """
        `Retrieve one or more unit <https://developer.cognite.com/api#tag/Units/operation/byIdsUnits>`_

        Args:
            external_id (str | SequenceNotStr[str]): External ID or list of external IDs
            ignore_unknown_ids (bool): Ignore external IDs that are not found rather than throw an exception.

        Returns:
            Unit | UnitList | None: If a single external ID is specified: the requested unit, or None if it does not exist. If several external IDs are specified: the requested units.

        Examples:

            Retrive unit 'temperature:deg_c':

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.units.retrieve('temperature:deg_c')

            Retrive units 'temperature:deg_c' and 'pressure:bar':

                >>> res = client.units.retrieve(['temperature:deg_c', 'pressure:bar'])
        """
        return run_sync(
            self.__async_client.units.retrieve(external_id=external_id, ignore_unknown_ids=ignore_unknown_ids)
        )

    @overload
    def from_alias(
        self,
        alias: str,
        quantity: str | None = None,
        *,
        return_ambiguous: Literal[False] = False,
        return_closest_matches: Literal[False] = False,
    ) -> Unit: ...

    @overload
    def from_alias(
        self,
        alias: str,
        quantity: str | None = None,
        *,
        return_ambiguous: bool = False,
        return_closest_matches: bool = False,
    ) -> UnitList: ...

    def from_alias(
        self,
        alias: str,
        quantity: str | None = None,
        *,
        return_ambiguous: bool = False,
        return_closest_matches: bool = False,
    ) -> Unit | UnitList:
        """
        Look up a unit by alias, optionally for a given quantity. Aliases and quantities are case-sensitive.

        Note:
            When just ``alias`` is given (i.e. ``quantity`` is not specified), some aliases are ambiguous as they are used
            by several quantities, e.g. 'F' which can be both Farad (Capacitance) and Fahrenheit (Temperature). These raise
            a ValueError by default unless also ``return_ambiguous=True`` is passed, in which case all matching units are returned.

        Tip:
            You can use ``return_closest_matches=True`` to get the closest matching units if the lookup fails. Note that there
            may not be any close matches, in which case an empty UnitList is returned.

        Args:
            alias (str): Alias of the unit, like 'cmol / L' or 'meter per second'.
            quantity (str | None): Quantity of the unit, like 'Temperature' or 'Pressure'.
            return_ambiguous (bool): If False (default), when the alias is ambiguous (i.e. no quantity was given), raise a ValueError. If True, return the list of all matching units.
            return_closest_matches (bool): If False (default), when the lookup fails, raise a ValueError (default). If True, return the closest matching units (even if empty).

        Returns:
            Unit | UnitList: The unit if found, else a ValueError is raised. If one or both of ``return_ambiguous`` and ``return_closest_matches`` is passed as True, a UnitList may be returned.

        Examples:

                Look up a unit by alias only:

                    >>> from cognite.client import CogniteClient, AsyncCogniteClient
                    >>> client = CogniteClient()
                    >>> # async_client = AsyncCogniteClient()  # another option
                    >>> unit = client.units.from_alias('cmol / L')

                Look up ambiguous alias 'F' by passing quantity 'Temperature':

                    >>> unit = client.units.from_alias('F', 'Temperature')

                Search for the closest matching unit of 'kilo watt' (should be 'kilowatt'):

                    >>> unit_matches = client.units.from_alias("kilo watt", return_closest_matches=True)
        """
        return run_sync(
            self.__async_client.units.from_alias(
                alias=alias,
                quantity=quantity,
                return_ambiguous=return_ambiguous,
                return_closest_matches=return_closest_matches,
            )
        )

    def list(self) -> UnitList:
        """
        `List all supported units <https://developer.cognite.com/api#tag/Units/operation/listUnits>`_

        Returns:
            UnitList: List of units

        Examples:

            List all supported units in CDF:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.units.list()
        """
        return run_sync(self.__async_client.units.list())



"""
===============================================================================
dee7369c5fe919ffd146a7c16845acea
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from cognite.client._api.iam import ComparableCapability
from cognite.client._sync_api.iam.groups import SyncGroupsAPI
from cognite.client._sync_api.iam.security_categories import SyncSecurityCategoriesAPI
from cognite.client._sync_api.iam.sessions import SyncSessionsAPI
from cognite.client._sync_api.iam.token import SyncTokenAPI
from cognite.client._sync_api.org_apis.principals import SyncPrincipalsAPI
from cognite.client._sync_api.user_profiles import SyncUserProfilesAPI
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.capabilities import Capability
from cognite.client.utils._async_helpers import run_sync

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class SyncIAMAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client
        self.groups = SyncGroupsAPI(async_client)
        self.security_categories = SyncSecurityCategoriesAPI(async_client)
        self.sessions = SyncSessionsAPI(async_client)
        self.user_profiles = SyncUserProfilesAPI(async_client)
        self.principals = SyncPrincipalsAPI(async_client)
        self.token = SyncTokenAPI(async_client)

    def compare_capabilities(
        self,
        existing_capabilities: ComparableCapability,
        desired_capabilities: ComparableCapability,
        project: str | None = None,
    ) -> list[Capability]:
        """
        Helper method to compare capabilities across two groups (of capabilities) to find which are missing from the first.

        Note:
            Capabilities that are no longer in use by the API will be ignored. These have names prefixed with `Legacy` and
            all inherit from the base class `LegacyCapability`. If you want to check for these, you must do so manually.

        Tip:
            If you just want to check against your existing capabilities, you may use the helper method
            ``client.iam.verify_capabilities`` instead.

        Args:
            existing_capabilities (ComparableCapability): List of existing capabilities.
            desired_capabilities (ComparableCapability): List of wanted capabilities to check against existing.
            project (str | None): If a ProjectCapability or ProjectCapabilityList is passed, we need to know which CDF project
                to pull capabilities from (existing might be from several). If project is not passed, and ProjectCapabilityList
                is used, it will be inferred from the AsyncCogniteClient used to call retrieve it via token/inspect.

        Returns:
            list[Capability]: A flattened list of the missing capabilities, meaning they each have exactly 1 action, 1 scope, 1 id etc.

        Examples:

            Ensure that a user's groups grant access to read- and write for assets in all scope,
            and events write, scoped to a specific dataset with id=123:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.capabilities import AssetsAcl, EventsAcl
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> my_groups = client.iam.groups.list(all=False)
                >>> to_check = [
                ...     AssetsAcl(
                ...         actions=[AssetsAcl.Action.Read, AssetsAcl.Action.Write],
                ...         scope=AssetsAcl.Scope.All()),
                ...     EventsAcl(
                ...         actions=[EventsAcl.Action.Write],
                ...         scope=EventsAcl.Scope.DataSet([123]),
                ... )]
                >>> missing = client.iam.compare_capabilities(
                ...     existing_capabilities=my_groups,
                ...     desired_capabilities=to_check)
                >>> if missing:
                ...     pass  # do something

            Capabilities can also be passed as dictionaries:

                >>> to_check = [
                ...     {'assetsAcl': {'actions': ['READ', 'WRITE'], 'scope': {'all': {}}}},
                ...     {'eventsAcl': {'actions': ['WRITE'], 'scope': {'datasetScope': {'ids': [123]}}}},
                ... ]
                >>> missing = client.iam.compare_capabilities(
                ...     existing_capabilities=my_groups,
                ...     desired_capabilities=to_check)

            You may also load capabilities from a dict-representation directly into ACLs (access-control list)
            by using ``Capability.load``. This will also ensure that the capabilities are valid.

                >>> from cognite.client.data_classes.capabilities import Capability
                >>> acls = [Capability.load(cap) for cap in to_check]
        """
        return self.__async_client.iam.compare_capabilities(
            existing_capabilities=existing_capabilities, desired_capabilities=desired_capabilities, project=project
        )

    def verify_capabilities(self, desired_capabilities: ComparableCapability) -> list[Capability]:
        """
        Helper method to compare your current capabilities with a set of desired capabilities and return any missing.

        Args:
            desired_capabilities (ComparableCapability): List of desired capabilities to check against existing.

        Returns:
            list[Capability]: A flattened list of the missing capabilities, meaning they each have exactly 1 action, 1 scope, 1 id etc.

        Examples:

            Ensure that the user's credentials have access to read- and write assets in all scope,
            and write events scoped to a specific dataset with id=123:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.capabilities import AssetsAcl, EventsAcl
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> to_check = [
                ...     AssetsAcl(
                ...         actions=[AssetsAcl.Action.Read, AssetsAcl.Action.Write],
                ...         scope=AssetsAcl.Scope.All()),
                ...     EventsAcl(
                ...         actions=[EventsAcl.Action.Write],
                ...         scope=EventsAcl.Scope.DataSet([123]),
                ... )]
                >>> if missing := client.iam.verify_capabilities(to_check):
                ...     pass  # do something

            Capabilities can also be passed as dictionaries:

                >>> to_check = [
                ...     {'assetsAcl': {'actions': ['READ', 'WRITE'], 'scope': {'all': {}}}},
                ...     {'eventsAcl': {'actions': ['WRITE'], 'scope': {'datasetScope': {'ids': [123]}}}},
                ... ]
                >>> missing = client.iam.verify_capabilities(to_check)

            You may also load capabilities from a dict-representation directly into ACLs (access-control list)
            by using ``Capability.load``. This will also ensure that the capabilities are valid.

                >>> from cognite.client.data_classes.capabilities import Capability
                >>> acls = [Capability.load(cap) for cap in to_check]
        """
        return run_sync(self.__async_client.iam.verify_capabilities(desired_capabilities=desired_capabilities))



from __future__ import annotations

import typing
from collections.abc import AsyncIterator, Iterator
from datetime import timedelta
from typing import Any, Final

from httpx import URL, Cookies, Headers, HTTPStatusError
from httpx import Request as HttpxRequest
from httpx import Response as HttpxResponse


class CogniteHTTPResponse:
    """
    A wrapper class (currently around `httpx.Response`) to isolate the SDK from the
    underlying HTTP library's public interface.
    """

    # Every single response is wrapped in this class, so squeeze out every bit of perf:
    __slots__ = ("_json_cache", "_response")
    _json_cache: Any

    def __init__(self, response: HttpxResponse) -> None:
        self._response: Final[HttpxResponse] = response

    @property
    def httpx_response(self) -> HttpxResponse:
        """
        Direct access to the Response object from the underlying http library (currently httpx).

        Disclaimer: Usage is neither backwards- nor forwards-compatible.
        """
        return self._response

    @property
    def status_code(self) -> int:
        return self._response.status_code

    @property
    def reason_phrase(self) -> str:
        return self._response.reason_phrase

    @property
    def http_version(self) -> str:
        return self._response.http_version

    @property
    def url(self) -> URL:
        return self._response.url

    @property
    def headers(self) -> Headers:
        return self._response.headers

    @property
    def cookies(self) -> Cookies:
        # Note: httpx.Response.cookies returns httpx.Cookies, which is a wrapper around SimpleCookie
        return self._response.cookies

    @property
    def history(self) -> list[CogniteHTTPResponse]:
        return [CogniteHTTPResponse(r) for r in self._response.history]

    @property
    def request(self) -> HttpxRequest:
        return self._response.request

    @property
    def content(self) -> bytes:
        return self._response.content

    @property
    def text(self) -> str:
        return self._response.text

    @property
    def encoding(self) -> str | None:
        return self._response.encoding

    @encoding.setter
    def encoding(self, value: str) -> None:
        self._response.encoding = value

    @property
    def is_success(self) -> bool:
        return self._response.is_success

    @property
    def is_error(self) -> bool:
        return self._response.is_error

    @property
    def elapsed(self) -> timedelta:
        return self._response.elapsed

    @property
    def next_request(self) -> HttpxRequest | None:
        return self._response.next_request

    @property
    def stream(self) -> typing.Any:
        # 'stream' typing is tricky; we leave it at Any as users should consume from the stream
        # using the provided methods like iter_bytes, iter_lines, iter_text, etc.
        return self._response.stream

    @property
    def num_bytes_downloaded(self) -> int:
        return self._response.num_bytes_downloaded

    def json(self) -> Any:
        try:
            return self._json_cache
        except AttributeError:
            self._json_cache = self._response.json()
            return self._json_cache

    def raise_for_status(self) -> CogniteHTTPResponse:
        """
        Raises a CogniteHTTPStatusError if the response status code is 4xx or 5xx.
        If successful (2xx), returns the response object (self) for chaining.
        """
        from cognite.client.exceptions import CogniteHTTPStatusError

        try:
            self._response.raise_for_status()
            return self
        except HTTPStatusError:
            raise CogniteHTTPStatusError(self.status_code, request=self.request, response=self) from None

    def iter_bytes(self, chunk_size: int | None = None) -> Iterator[bytes]:
        return self._response.iter_bytes(chunk_size)

    def read(self) -> bytes:
        return self._response.read()

    def close(self) -> None:
        self._response.close()

    async def aiter_bytes(self, chunk_size: int | None = None) -> AsyncIterator[bytes]:
        async for chunk in self._response.aiter_bytes(chunk_size):
            yield chunk

    async def aiter_text(self, chunk_size: int | None = None) -> AsyncIterator[str]:
        async for text in self._response.aiter_text(chunk_size):
            yield text

    async def aiter_lines(self) -> AsyncIterator[str]:
        async for line in self._response.aiter_lines():
            yield line

    async def aiter_raw(self, chunk_size: int | None = None) -> AsyncIterator[bytes]:
        async for chunk in self._response.aiter_raw(chunk_size):
            yield chunk

    async def aread(self) -> bytes:
        return await self._response.aread()

    async def aclose(self) -> None:
        await self._response.aclose()



"""
===============================================================================
ebf94ce24437dd81bab208403099ec24
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from cognite.client import AsyncCogniteClient
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.units import UnitSystemList
from cognite.client.utils._async_helpers import run_sync


class SyncUnitSystemAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    def list(self) -> UnitSystemList:
        """
        `List all supported unit systems <https://developer.cognite.com/api#tag/Unit-Systems/operation/listUnitSystems>`_

        Returns:
            UnitSystemList: List of unit systems

        Examples:

            List all supported unit systems in CDF:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.units.systems.list()
        """
        return run_sync(self.__async_client.units.systems.list())



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, TypeVar, cast

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteFilter,
    CogniteObject,
    CogniteResourceList,
    UnknownCogniteObject,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.data_modeling._validation import validate_data_modeling_identifier
from cognite.client.data_classes.data_modeling.containers import PropertyConstraintState
from cognite.client.data_classes.data_modeling.core import DataModelingSchemaResource
from cognite.client.data_classes.data_modeling.data_types import (
    DirectRelation,
    DirectRelationReference,
    PropertyType,
)
from cognite.client.data_classes.data_modeling.ids import ContainerId, PropertyId, ViewId
from cognite.client.data_classes.filters import Filter
from cognite.client.utils._text import convert_all_keys_to_camel_case_recursive, to_camel_case, to_snake_case

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class ViewCore(DataModelingSchemaResource["ViewApply"], ABC):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: str,
        description: str | None,
        name: str | None,
        filter: Filter | None,
        implements: list[ViewId] | None,
    ) -> None:
        super().__init__(space=space, external_id=external_id, name=name, description=description)
        self.filter = filter
        self.implements: list[ViewId] = implements or []
        self.version = version

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)

        if self.implements:
            output["implements"] = [v.dump(camel_case) for v in self.implements]
        if self.filter is not None:
            output["filter"] = self.filter.dump()

        return output

    def as_id(self) -> ViewId:
        return ViewId(
            space=self.space,
            external_id=self.external_id,
            version=self.version,
        )

    def as_property_ref(self, property: str) -> tuple[str, str, str]:
        return self.as_id().as_property_ref(property)


class ViewApply(ViewCore):
    """A group of properties. Write only version.

    Args:
        space (str): The workspace for the view, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the view.
        version (str): DMS version.
        description (str | None): Textual description of the view
        name (str | None): Human readable name for the view.
        filter (Filter | None): A filter Domain Specific Language (DSL) used to create advanced filter queries.
        implements (list[ViewId] | None): References to the views from where this view will inherit properties and edges.
        properties (dict[str, ViewPropertyApply] | None): No description.

    .. note::
        The order of elements (i.e., `ViewId`) in :code:`implements` matters, as it indicates priority on how to handle
        collisions of same properties from different views.
        See docs on
        `implemented property conflicts <https://docs.cognite.com/cdf/dm/dm_concepts/dm_containers_views_datamodels/#implemented-property-conflicts-and-precedence>`_
        for more details.

    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: str,
        description: str | None = None,
        name: str | None = None,
        filter: Filter | None = None,
        implements: list[ViewId] | None = None,
        properties: dict[str, ViewPropertyApply] | None = None,
    ) -> None:
        validate_data_modeling_identifier(space, external_id)
        super().__init__(space, external_id, version, description, name, filter, implements)
        self.properties = properties

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        properties = (
            {k: ViewPropertyApply.load(v) for k, v in resource["properties"].items()}
            if "properties" in resource
            else None
        )
        implements = [ViewId.load(v) for v in resource["implements"]] if "implements" in resource else None
        filter = Filter.load(resource["filter"]) if "filter" in resource else None
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            version=resource["version"],
            description=resource.get("description"),
            name=resource.get("name"),
            filter=filter,
            implements=implements,
            properties=properties,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if "properties" in output:
            output["properties"] = {k: v.dump(camel_case) for k, v in output["properties"].items()}

        return output

    def as_write(self) -> ViewApply:
        """Returns this ViewApply instance."""
        return self

    def referenced_containers(self) -> set[ContainerId]:
        """Helper function to get the set of containers referenced by this view.

        Returns:
            set[ContainerId]: The set of containers referenced by this view.
        """
        referenced_containers = set()
        for prop in (self.properties or {}).values():
            if isinstance(prop, MappedPropertyApply):
                referenced_containers.add(prop.container)
        return referenced_containers


class View(ViewCore):
    """A group of properties. Read only version.

    Args:
        space (str): The workspace for the view, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the view.
        version (str): DMS version.
        properties (dict[str, ViewProperty]): View with included properties and expected edges, indexed by a unique space-local identifier.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        description (str | None): Textual description of the view
        name (str | None): Human readable name for the view.
        filter (Filter | None): A filter Domain Specific Language (DSL) used to create advanced filter queries.
        implements (list[ViewId] | None): References to the views from where this view will inherit properties and edges.
        writable (bool): Whether the view supports write operations.
        used_for (Literal['node', 'edge', 'all']): Does this view apply to nodes, edges or both.
        is_global (bool): Whether this is a global view.
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: str,
        properties: dict[str, ViewProperty],
        last_updated_time: int,
        created_time: int,
        description: str | None,
        name: str | None,
        filter: Filter | None,
        implements: list[ViewId] | None,
        writable: bool,
        used_for: Literal["node", "edge", "all"],
        is_global: bool,
    ) -> None:
        super().__init__(
            space,
            external_id,
            version,
            description,
            name,
            filter,
            implements,
        )
        self.writable = writable
        self.used_for = used_for
        self.is_global = is_global
        self.properties = properties
        self.last_updated_time = last_updated_time
        self.created_time = created_time

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> View:
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            version=resource["version"],
            description=resource.get("description"),
            name=resource.get("name"),
            last_updated_time=resource["lastUpdatedTime"],
            created_time=resource["createdTime"],
            filter=Filter.load(resource["filter"]) if "filter" in resource else None,
            implements=[ViewId.load(v) for v in resource["implements"]] if "implements" in resource else None,
            writable=resource["writable"],
            used_for=resource["usedFor"],
            is_global=resource["isGlobal"],
            properties={k: ViewProperty.load(v) for k, v in resource.get("properties", {}).items()},
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if "properties" in output:
            output["properties"] = {k: v.dump(camel_case) for k, v in output["properties"].items()}

        return output

    def as_apply(self) -> ViewApply:
        """Convert to a view applies.

        Returns:
            ViewApply: The view apply.
        """
        properties: dict[str, ViewPropertyApply] | None = None
        if self.properties:
            for k, v in self.properties.items():
                if isinstance(
                    v,
                    (
                        MappedProperty,
                        SingleEdgeConnection,
                        MultiEdgeConnection,
                        SingleReverseDirectRelation,
                        MultiReverseDirectRelation,
                    ),
                ):
                    if properties is None:
                        properties = {}
                    properties[k] = v.as_apply()
                else:
                    raise NotImplementedError(f"Unsupported conversion to apply for property type {type(v)}")

        return ViewApply(
            space=self.space,
            external_id=self.external_id,
            version=self.version,
            description=self.description,
            name=self.name,
            filter=self.filter,
            implements=self.implements,
            properties=properties,
        )

    def as_write(self) -> ViewApply:
        return self.as_apply()

    def referenced_containers(self) -> set[ContainerId]:
        """Helper function to get the set of containers referenced by this view.

        Returns:
            set[ContainerId]: The set of containers referenced by this view.
        """
        referenced_containers = set()
        for prop in self.properties.values():
            if isinstance(prop, MappedProperty):
                referenced_containers.add(prop.container)
        return referenced_containers


class ViewApplyList(CogniteResourceList[ViewApply]):
    _RESOURCE = ViewApply

    def as_ids(self) -> list[ViewId]:
        """Returns the list of ViewIds

        Returns:
            list[ViewId]: The list of ViewIds
        """
        return [v.as_id() for v in self]

    def referenced_containers(self) -> set[ContainerId]:
        """Helper function to get the set of containers referenced by this view.

        Returns:
            set[ContainerId]: The set of containers referenced by this view.
        """
        referenced_containers = set()
        for view in self:
            referenced_containers.update(view.referenced_containers())
        return referenced_containers


class ViewList(WriteableCogniteResourceList[ViewApply, View]):
    _RESOURCE = View

    def as_apply(self) -> ViewApplyList:
        """Convert to a view an apply list.

        Returns:
            ViewApplyList: The view apply list.
        """
        return ViewApplyList(resources=[v.as_apply() for v in self])

    def as_ids(self) -> list[ViewId]:
        """Returns the list of ViewIds

        Returns:
            list[ViewId]: The list of ViewIds
        """
        return [v.as_id() for v in self]

    def as_write(self) -> ViewApplyList:
        return self.as_apply()

    def referenced_containers(self) -> set[ContainerId]:
        """Helper function to get the set of containers referenced by this view.

        Returns:
            set[ContainerId]: The set of containers referenced by this view.
        """
        referenced_containers = set()
        for view in self:
            referenced_containers.update(view.referenced_containers())
        return referenced_containers


class ViewFilter(CogniteFilter):
    """Represent the filer arguments for the list endpoint.

    Args:
        space (str | None): The space to query
        include_inherited_properties (bool): Whether to include properties inherited from views this view implements.
        all_versions (bool): Whether to return all versions. If false, only the newest version is returned, which is determined based on the 'createdTime' field.
        include_global (bool): Whether to include global views.
    """

    def __init__(
        self,
        space: str | None = None,
        include_inherited_properties: bool = True,
        all_versions: bool = False,
        include_global: bool = False,
    ) -> None:
        self.space = space
        self.include_inherited_properties = include_inherited_properties
        self.all_versions = all_versions
        self.include_global = include_global


class ViewProperty(CogniteObject, ABC):
    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ViewProperty:
        if "connectionType" in resource:
            return ConnectionDefinition.load(resource)
        elif "direction" in resource:
            raise ValueError("Connection Definition is missing field 'connectionType'")
        else:
            return MappedProperty.load(resource)

    @abstractmethod
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        raise NotImplementedError


class ViewPropertyApply(CogniteObject, ABC):
    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ViewPropertyApply:
        if "connectionType" in resource:
            return ConnectionDefinitionApply.load(resource)
        elif "direction" in resource:
            raise ValueError("Connection Definition is missing field 'connectionType'")
        else:
            return MappedPropertyApply.load(resource)

    @abstractmethod
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        raise NotImplementedError


@dataclass
class MappedPropertyApply(ViewPropertyApply):
    container: ContainerId
    container_property_identifier: str
    name: str | None = None
    description: str | None = None
    source: ViewId | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            container=ContainerId.load(resource["container"]),
            container_property_identifier=resource["containerPropertyIdentifier"],
            name=resource.get("name"),
            description=resource.get("description"),
            source=ViewId.load(resource["source"]) if "source" in resource else None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        key = "containerPropertyIdentifier" if camel_case else "container_property_identifier"
        output: dict[str, Any] = {
            "container": self.container.dump(camel_case, include_type=True),
            key: self.container_property_identifier,
        }
        if self.name is not None:
            output["name"] = self.name
        if self.description is not None:
            output["description"] = self.description
        if self.source is not None:
            output["source"] = self.source.dump(camel_case, include_type=True)

        return output


@dataclass
class MappedProperty(ViewProperty):
    container: ContainerId
    container_property_identifier: str
    type: PropertyType
    nullable: bool
    immutable: bool
    auto_increment: bool
    source: ViewId | None = None
    default_value: str | int | dict | None = None
    name: str | None = None
    description: str | None = None
    constraint_state: PropertyConstraintState | None = field(default=None, init=False)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        type_ = resource["type"]
        source = type_.get("source", None) or resource.get("source")

        prop = cls(
            container=ContainerId.load(resource["container"]),
            container_property_identifier=resource["containerPropertyIdentifier"],
            type=PropertyType.load({k: v for k, v in type_.items() if k != "source"}),
            nullable=resource["nullable"],
            immutable=resource["immutable"],
            auto_increment=resource["autoIncrement"],
            source=ViewId.load(source) if source else None,
            default_value=resource.get("defaultValue"),
            name=resource.get("name"),
            description=resource.get("description"),
        )
        if (constraint_state := resource.get("constraintState")) is not None:
            object.__setattr__(prop, "constraint_state", PropertyConstraintState._load(constraint_state))
        return prop

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {}
        output["container"] = self.container.dump(camel_case, include_type=False)
        output["containerPropertyIdentifier" if camel_case else "container_property_identifier"] = (
            self.container_property_identifier
        )
        if self.source:
            output["source"] = self.source.dump(camel_case, include_type=False)
        output["type"] = self.type.dump(camel_case)
        output["immutable"] = self.immutable
        for key in ["nullable", "auto_increment", "name", "default_value", "description"]:
            if (value := getattr(self, key)) is not None:
                output[to_camel_case(key) if camel_case else key] = value
        if self.constraint_state is not None:
            output["constraintState" if camel_case else "constraint_state"] = self.constraint_state.dump(camel_case)
        if self.source and isinstance(self.type, DirectRelation):
            output["type"]["source"] = output.pop("source", None)
        return output

    def as_apply(self) -> MappedPropertyApply:
        return MappedPropertyApply(
            container=self.container,
            container_property_identifier=self.container_property_identifier,
            name=self.name,
            description=self.description,
            source=self.source,
        )


@dataclass
class ConnectionDefinition(ViewProperty, ABC):
    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ConnectionDefinition:
        if "connectionType" not in resource:
            raise ValueError(f"{cls.__name__} must have a connectionType")
        connection_type = to_snake_case(resource["connectionType"])

        if connection_type == "single_edge_connection":
            return SingleEdgeConnection.load(resource)
        if connection_type == "multi_edge_connection":
            return MultiEdgeConnection.load(resource)
        if connection_type == "single_reverse_direct_relation":
            return SingleReverseDirectRelation.load(resource)
        if connection_type == "multi_reverse_direct_relation":
            return MultiReverseDirectRelation.load(resource)

        return cast(Self, UnknownCogniteObject(resource))

    @abstractmethod
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        raise NotImplementedError


@dataclass
class EdgeConnection(ConnectionDefinition, ABC):
    """Describes the edge(s) that are likely to exist to aid in discovery and documentation of the view.
    A listed edge is not required. i.e. It does not have to exist when included in this list.
    A connection has a max distance of one hop.

    Args:
        type (DirectRelationReference): Reference to the node pointed to by the direct relation. The reference
            consists of a space and an external-id.
        source (ViewId): The target node(s) of this connection can be read through the view specified in 'source'.
        name (str | None): Readable property name.
        description (str | None): Description of the content and suggested use for this property.
        edge_source (ViewId | None): The edge(s) of this connection can be read through the view specified in
            'edgeSource'.
        direction (Literal["outwards", "inwards"]): The direction of the edge. The outward direction is used to
            indicate that the edge points from the source to the target. The inward direction is used to indicate
            that the edge points from the target to the source.
    """

    type: DirectRelationReference
    source: ViewId
    name: str | None
    description: str | None
    edge_source: ViewId | None
    direction: Literal["outwards", "inwards"]

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        instance = cls(
            type=DirectRelationReference.load(resource["type"]),
            source=ViewId.load(resource["source"]),
            name=resource.get("name"),
            description=resource.get("description"),
            edge_source=(edge_source := resource.get("edgeSource")) and ViewId.load(edge_source),
            direction=resource["direction"],
        )

        return instance

    @abstractmethod
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = asdict(self)
        if self.type:
            output["type"] = self.type.dump(camel_case)
        if self.source:
            output["source"] = self.source.dump(camel_case)
        if self.edge_source:
            output["edge_source"] = self.edge_source.dump(camel_case)

        return convert_all_keys_to_camel_case_recursive(output) if camel_case else output


@dataclass
class SingleEdgeConnection(EdgeConnection):
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = super().dump(camel_case)
        if camel_case:
            output["connectionType"] = "single_edge_connection"
        else:
            output["connection_type"] = "single_edge_connection"

        return output

    def as_apply(self) -> SingleEdgeConnectionApply:
        return SingleEdgeConnectionApply(
            type=self.type,
            source=self.source,
            name=self.name,
            description=self.description,
            edge_source=self.edge_source,
            direction=self.direction,
        )


@dataclass
class MultiEdgeConnection(EdgeConnection):
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = super().dump(camel_case)
        if camel_case:
            output["connectionType"] = "multi_edge_connection"
        else:
            output["connection_type"] = "multi_edge_connection"

        return output

    def as_apply(self) -> MultiEdgeConnectionApply:
        return MultiEdgeConnectionApply(
            type=self.type,
            source=self.source,
            name=self.name,
            description=self.description,
            edge_source=self.edge_source,
            direction=self.direction,
        )


SingleHopConnectionDefinition: TypeAlias = MultiEdgeConnection


@dataclass
class ReverseDirectRelation(ConnectionDefinition, ABC):
    """Describes the direct relation(s) pointing to instances read through this view. This connection type is used to
    aid in discovery and documentation of the view

    It is called 'ReverseDirectRelationConnection' in the API spec.

    Args:
        source (ViewId): The node(s) containing the direct relation property can be read through
            the view specified in 'source'.
        through (PropertyId): The view or container of the node containing the direct relation property.
        name (str | None): Readable property name.
        description (str | None): Description of the content and suggested use for this property.

    """

    source: ViewId
    through: PropertyId
    name: str | None = None
    description: str | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            source=ViewId.load(resource["source"]),
            through=PropertyId.load(resource["through"]),
            name=resource.get("name"),
            description=resource.get("description"),
        )

    @abstractmethod
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "source": self.source.dump(camel_case),
            "through": self.through.dump(camel_case),
            "name": self.name,
            "description": self.description,
        }


@dataclass
class SingleReverseDirectRelation(ReverseDirectRelation):
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = super().dump(camel_case)
        if camel_case:
            output["connectionType"] = "single_reverse_direct_relation"
        else:
            output["connection_type"] = "single_reverse_direct_relation"

        return output

    def as_apply(self) -> SingleReverseDirectRelationApply:
        return SingleReverseDirectRelationApply(
            source=self.source,
            through=self.through,
            name=self.name,
            description=self.description,
        )


@dataclass
class MultiReverseDirectRelation(ReverseDirectRelation):
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = super().dump(camel_case)
        if camel_case:
            output["connectionType"] = "multi_reverse_direct_relation"
        else:
            output["connection_type"] = "multi_reverse_direct_relation"

        return output

    def as_apply(self) -> MultiReverseDirectRelationApply:
        return MultiReverseDirectRelationApply(
            source=self.source,
            through=self.through,
            name=self.name,
            description=self.description,
        )


@dataclass
class ConnectionDefinitionApply(ViewPropertyApply, ABC):
    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        if "connectionType" not in resource:
            raise ValueError(f"{cls.__name__} must have a connectionType")
        connection_type = to_snake_case(resource["connectionType"])

        if connection_type == "single_edge_connection":
            return cast(Self, SingleEdgeConnectionApply.load(resource))
        if connection_type == "multi_edge_connection":
            return cast(Self, MultiEdgeConnectionApply.load(resource))
        if connection_type == "single_reverse_direct_relation":
            return cast(Self, SingleReverseDirectRelationApply.load(resource))
        if connection_type == "multi_reverse_direct_relation":
            return cast(Self, MultiReverseDirectRelationApply.load(resource))
        return cast(Self, UnknownCogniteObject(resource))

    @abstractmethod
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        raise NotImplementedError


T_ConnectionDefinitionApply = TypeVar("T_ConnectionDefinitionApply", bound=ConnectionDefinitionApply)


@dataclass
class EdgeConnectionApply(ConnectionDefinitionApply, ABC):
    """Describes the edge(s) that are likely to exist to aid in discovery and documentation of the view.
    A listed edge is not required. i.e. It does not have to exist when included in this list.
    A connection has a max distance of one hop.

    It is called 'EdgeConnection' in the API spec.

    Args:
        type (DirectRelationReference): Reference to the node pointed to by the direct relation. The reference
            consists of a space and an external-id.
        source (ViewId): The target node(s) of this connection can be read through the view specified in 'source'.
        name (str | None): Readable property name.
        description (str | None): Description of the content and suggested use for this property.
        edge_source (ViewId | None): The edge(s) of this connection can be read through the view specified in
            'edgeSource'.
        direction (Literal["outwards", "inwards"]): The direction of the edge. The outward direction is used to
            indicate that the edge points from the source to the target. The inward direction is used to indicate
            that the edge points from the target to the source.
    """

    type: DirectRelationReference
    source: ViewId
    name: str | None = None
    description: str | None = None
    edge_source: ViewId | None = None
    direction: Literal["outwards", "inwards"] = "outwards"

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        instance = cls(
            type=DirectRelationReference.load(resource["type"]),
            source=ViewId.load(resource["source"]),
            name=resource.get("name"),
            description=resource.get("description"),
            edge_source=(edge_source := resource.get("edgeSource")) and ViewId.load(edge_source),
        )
        if "direction" in resource:
            instance.direction = resource["direction"]
        return instance

    @abstractmethod
    def dump(self, camel_case: bool = True) -> dict:
        output: dict[str, Any] = {
            "type": self.type.dump(camel_case),
            "source": self.source.dump(camel_case, include_type=True),
            "direction": self.direction,
        }
        if self.name is not None:
            output["name"] = self.name
        if self.description is not None:
            output["description"] = self.description
        if self.edge_source is not None:
            output[("edgeSource" if camel_case else "edge_source")] = self.edge_source.dump(
                camel_case, include_type=True
            )

        return output


@dataclass
class SingleEdgeConnectionApply(EdgeConnectionApply):
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = super().dump(camel_case)
        if camel_case:
            output["connectionType"] = "single_edge_connection"
        else:
            output["connection_type"] = "single_edge_connection"

        return output


@dataclass
class MultiEdgeConnectionApply(EdgeConnectionApply):
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = super().dump(camel_case)
        if camel_case:
            output["connectionType"] = "multi_edge_connection"
        else:
            output["connection_type"] = "multi_edge_connection"

        return output


SingleHopConnectionDefinitionApply: TypeAlias = MultiEdgeConnectionApply


@dataclass
class ReverseDirectRelationApply(ConnectionDefinitionApply, ABC):
    """Describes the direct relation(s) pointing to instances read through this view. This connection type is used to
    aid in discovery and documentation of the view.

    It is called 'ReverseDirectRelationConnection' in the API spec.

    Args:
        source (ViewId): The node(s) containing the direct relation property can be read through
            the view specified in 'source'.
        through (PropertyId): The view or container of the node containing the direct relation property.
        name (str | None): Readable property name.
        description (str | None): Description of the content and suggested use for this property.

    """

    source: ViewId
    through: PropertyId
    name: str | None = None
    description: str | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        instance = cls(
            source=ViewId.load(resource["source"]),
            through=PropertyId.load(resource["through"]),
        )
        if "name" in resource:
            instance.name = resource["name"]
        if "description" in resource:
            instance.description = resource["description"]

        return instance

    @abstractmethod
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {
            "source": self.source.dump(camel_case, include_type=True),
            "through": self.through.dump(camel_case),
        }
        if self.name is not None:
            output["name"] = self.name
        if self.description is not None:
            output["description"] = self.description

        return output


@dataclass
class SingleReverseDirectRelationApply(ReverseDirectRelationApply):
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = super().dump(camel_case)
        if camel_case:
            output["connectionType"] = "single_reverse_direct_relation"
        else:
            output["connection_type"] = "single_reverse_direct_relation"

        return output


@dataclass
class MultiReverseDirectRelationApply(ReverseDirectRelationApply):
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = super().dump(camel_case)
        if camel_case:
            output["connectionType"] = "multi_reverse_direct_relation"
        else:
            output["connection_type"] = "multi_reverse_direct_relation"

        return output



from __future__ import annotations

import os
import warnings
from collections.abc import AsyncIterable, AsyncIterator
from io import BufferedReader, BytesIO, StringIO, TextIOBase, UnsupportedOperation
from typing import Any, BinaryIO


def prepare_content_for_upload(
    content: str | bytes | BinaryIO | AsyncIterator[bytes],
) -> tuple[int | None, AsyncFileChunker | bytes | AsyncIterator[bytes]]:
    # Note: future dev, please no "hasattr read" checks here, we are strict on what we accept
    match content:
        case BufferedReader() | BytesIO():
            file_size = peek_filelike_length(content)
            content = AsyncFileChunker(content)
        case bytes():
            file_size = len(content)
        case str():
            content = content.encode("utf-8")
            file_size = len(content)
        case AsyncIterable():
            file_size = None
        case StringIO():
            raise TypeError("File uploads using 'io.StringIO' is not supported, please use 'io.BytesIO' instead.")
        case TextIOBase():
            raise TypeError("The file to upload must be opened in binary mode, not text mode.")
        case _:
            raise TypeError(
                f"Attempted to upload unsupported content of type {type(content)}. Supported types are bytes, "
                "str (will be encoded using UTF-8), io.BytesIO, and file handles (must be opened in binary mode). "
                "Async iterators yielding bytes are also supported for certain cloud providers."
            )
    if file_size is None:
        # We allow this as it works flawlessly on GCP
        warnings.warn(
            "Could not determine the size of the upload content. This leads to using chunked transfer encoding which "
            "may fail for certain cloud providers like Azure or AWS.",
            RuntimeWarning,
        )
    return file_size, content


class AsyncFileChunker(AsyncIterator[bytes]):
    """
    An asynchronous iterator for reading a file in chunks. Needed because httpx does not support
    file handles in a way that doesn't involve HTTP multipart encoding (as opposed to requests).

    Args:
        file_handle (BinaryIO): An open file handle.
    """

    CHUNK_SIZE = 64 * 1024  # 64 KiB chunks by default, copying httpx default

    def __init__(self, file_handle: BinaryIO) -> None:
        from cognite.client import global_config

        self._file_handle = file_handle
        self._chunk_size = global_config.file_upload_chunk_size or self.CHUNK_SIZE

        # Read from beginning of file if possible:
        if hasattr(self._file_handle, "seek"):
            try:
                self._file_handle.seek(0)
            except UnsupportedOperation:
                pass

    def __aiter__(self) -> AsyncIterator[bytes]:
        return self

    async def __anext__(self) -> bytes:
        if chunk := self._file_handle.read(self._chunk_size):
            return chunk
        raise StopAsyncIteration


# Straight from httpx/_utils.py (comments removed) as it's currently not
# exposed - unlike requests and its super_len function:
def peek_filelike_length(stream: Any) -> int | None:
    """
    Given a file-like stream object, return its length in number of bytes
    without reading it into memory.
    """
    try:
        fd = stream.fileno()
        length = os.fstat(fd).st_size
    except (AttributeError, OSError):
        try:
            offset = stream.tell()
            length = stream.seek(0, os.SEEK_END)
            stream.seek(offset)
        except (AttributeError, OSError):
            return None
    return length
